# -*- coding: utf-8 -*-
"""
This module adds a way to declare external functions.

See numba.function_util on how to call them.
"""
from __future__ import print_function, division, absolute_import

from numba.minivect import minitypes

class ExternalFunction(object):
    _attributes = ('func_name', 'arg_types', 'return_type', 'is_vararg',
                   'check_pyerr_occurred')
    func_name = None
    arg_types = None
    return_type = None
    is_vararg = False

    badval = None
    goodval = None
    exc_type = None
    exc_msg = None
    exc_args = None

    check_pyerr_occurred = False

    def __init__(self, return_type=None, arg_types=None, **kwargs):
        # Add positional arguments to keyword arguments
        if return_type is not None:
            kwargs['return_type'] = return_type
        if arg_types is not None:
            kwargs['arg_types'] = arg_types

        # Process keyword arguments
        if __debug__:
            # Only accept keyword arguments defined _attributes
            for k, v in kwargs.items():
                if k not in self._attributes:
                    raise TypeError("Invalid keyword arg %s -> %s" % (k, v))
        vars(self).update(kwargs)

    @property
    def signature(self):
        return minitypes.FunctionType(return_type=self.return_type,
                                      args=self.arg_types,
                                      is_vararg=self.is_vararg)

    @property
    def name(self):
        if self.func_name is None:
            return type(self).__name__
        else:
            return self.func_name

    def declare_lfunc(self, context, llvm_module):
        lfunc_type = self.signature.to_llvm(context)
        lfunc = llvm_module.get_or_insert_function(lfunc_type, name=self.name)
        return lfunc

class ExternalLibrary(object):
    def __init__(self, context):
        self._context = context
        # (name) -> (external function instance)
        self._functions = {}

    def add(self, extfn):
        if __debug__:
            # Sentry for duplicated external function name
            if extfn.name in self._functions:
                raise NameError("Duplicated external function: %s" % extfn.name)
        self._functions[extfn.name] = extfn

    def get(self, name):
        return self._functions[name]

    def __contains__(self, name):
        return name in self._functions

    def declare(self, module, name, arg_types=(), return_type=None):
        extfn = self._functions[name] # raises KeyError if name is not found

        if arg_types and return_type:
            if (extfn.arg_types != arg_types
                and extfn.return_type != return_type):
                raise TypeError("Signature mismatch")

        sig = extfn.signature
        lfunc_type = sig.to_llvm(self._context)
        return sig, module.get_or_insert_function(lfunc_type, extfn.name)






# -*- coding: utf-8 -*-
"""
This module provides a minimal type system, and ways to promote types, as
well as ways to convert to an LLVM type system. A set of predefined types are
defined. Types may be sliced to turn them into array types, in the same way
as the memoryview syntax.

>>> char
char
>>> int8[:, :, :]
int8[:, :, :]
>>> int8.signed
True
>>> uint8
uint8
>>> uint8.signed
False

>>> char.pointer()
char *
>>> int_[:, ::1]
int[:, ::1]
>>> int_[::1, :]
int[::1, :]
>>> double[:, ::1, :]
Traceback (most recent call last):
   ...
InvalidTypeSpecification: Step may only be provided once, and only in the first or last dimension.
"""
from __future__ import print_function, division, absolute_import

__all__ = ['Py_ssize_t', 'void', 'char', 'uchar', 'short', 'ushort',
           'int_', 'uint', 'long_', 'ulong', 'longlong', 'ulonglong',
           'size_t', 'npy_intp', 'c_string_type', 'bool_', 'object_',
           'float_', 'double', 'longdouble', 'float32', 'float64', 'float128',
           'int8', 'int16', 'int32', 'int64', 'uint8', 'uint16', 'uint32', 'uint64',
           'complex64', 'complex128', 'complex256', 'struct', 'Py_uintptr_t']

import sys
import math
import copy
import struct as struct_
import types
import textwrap

from . import miniutils
from . import minierror

from .miniutils import *
from .miniutils import have_ctypes, ctypes

_plat_bits = struct_.calcsize('@P') * 8

if struct_.pack('i', 1)[0] == '\1':
    nbo = '<' # little endian
else:
    nbo = '>' # big endian

class TypeMapper(object):
    """
    Maps foreign types to minitypes. Users of minivect should implement
    this and pass it to :py:class:`minivect.miniast.Context`.

    >>> import miniast
    >>> context = miniast.Context()
    >>> miniast.typemapper = TypeMapper(context)
    >>> tm = context.typemapper

    >>> tm.promote_types(int8, double)
    double
    >>> tm.promote_types(int8, uint8)
    uint8
    >>> tm.promote_types(int8, complex128)
    complex128
    >>> tm.promote_types(int8, object_)
    PyObject *

    >>> tm.promote_types(int64, float32)
    float32
    >>> tm.promote_types(int64, complex64)
    complex64
    >>> tm.promote_types(float32, float64)
    float64
    >>> tm.promote_types(float32, complex64)
    complex64
    >>> tm.promote_types(complex64, complex128)
    complex128
    >>> tm.promote_types(complex256, object_)
    PyObject *

    >>> tm.promote_types(float32.pointer(), Py_ssize_t)
    float32 *
    >>> tm.promote_types(float32.pointer(), Py_ssize_t)
    float32 *
    >>> tm.promote_types(float32.pointer(), uint8)
    float32 *

    >>> tm.promote_types(float32.pointer(), float64.pointer())
    Traceback (most recent call last):
        ...
    UnpromotableTypeError: (float32 *, float64 *)

    >>> tm.promote_types(float32[:, ::1], float32[:, ::1])
    float32[:, ::1]
    >>> tm.promote_types(float32[:, ::1], float64[:, ::1])
    float64[:, ::1]
    >>> tm.promote_types(float32[:, ::1], float64[::1, :])
    float64[:, :]
    >>> tm.promote_types(float32[:, :], complex128[:, :])
    complex128[:, :]
    >>> tm.promote_types(int_[:, :], object_[:, ::1])
    PyObject *[:, :]

    >>> tm.promote_types(npy_intp, float_)
    float
    >>> tm.promote_types(npy_intp, float32)
    float32

    >>> float32 == float_
    True
    >>> float64 == double
    True
    >>> float128 == longdouble
    True
    """

    def __init__(self, context):
        self.context = context

    def map_type(self, opaque_type):
        "Map a foreign type to a minitype"
        if opaque_type.is_int:
            return int_
        elif opaque_type.is_float:
            return float_
        elif opaque_type.is_double:
            return double
        elif opaque_type.is_pointer:
            return PointerType(self.map_type(opaque_type.base_type))
        elif opaque_type.is_py_ssize_t:
            return Py_ssize_t
        elif opaque_type.is_char:
            return char
        else:
            raise minierror.UnmappableTypeError(opaque_type)

    def to_llvm(self, type):
        "Return an LLVM type for the given type."
        raise NotImplementedError(type)

    def from_python(self, value):
        "Get a type from a python value"
        np = sys.modules.get('numpy', None)

        if isinstance(value, float):
            return double
        elif isinstance(value, bool):
            return bool_
        elif isinstance(value, (int, long)):
            if abs(value) < 1:
                bits = 0
            else:
                bits = math.ceil(math.log(abs(value), 2))

            if bits < 32:
                return int_
            elif bits < 64:
                return int64
            else:
                raise ValueError("Cannot represent %s as int32 or int64", value)
        elif isinstance(value, complex):
            return complex128
        elif isinstance(value, str):
            return c_string_type
        elif np and isinstance(value, np.ndarray):
            dtype = map_dtype(value.dtype)
            return ArrayType(dtype, value.ndim,
                             is_c_contig=value.flags['C_CONTIGUOUS'],
                             is_f_contig=value.flags['F_CONTIGUOUS'])
        else:
            return object_
            # raise minierror.UnmappableTypeError(type(value))

    def promote_numeric(self, type1, type2):
        "Promote two numeric types"
        type = max([type1, type2], key=lambda type: type.rank)
        if type1.kind != type2.kind:
            def itemsize(type):
                return type.itemsize // 2 if type.is_complex else type.itemsize

            size = max(itemsize(type1), itemsize(type2))
            if type.is_complex:
                type = find_type_of_size(size * 2, complextypes)
            elif type.is_float:
                type = find_type_of_size(size, floating)
            else:
                assert type.is_int
                type = find_type_of_size(size, integral)

        return type

    def promote_arrays(self, type1, type2):
        "Promote two array types in an expression to a new array type"
        equal_ndim = type1.ndim == type2.ndim
        return ArrayType(self.promote_types(type1.dtype, type2.dtype),
                         ndim=max((type1.ndim, type2.ndim)),
                         is_c_contig=(equal_ndim and type1.is_c_contig and
                                      type2.is_c_contig),
                         is_f_contig=(equal_ndim and type1.is_f_contig and
                                      type2.is_f_contig))

    def promote_types(self, type1, type2):
        "Promote two arbitrary types"
        string_types = c_string_type, char.pointer()
        if type1.is_pointer and type2.is_int_like:
            return type1
        elif type2.is_pointer and type2.is_int_like:
            return type2
        elif type1.is_object or type2.is_object:
            return object_
        elif type1.is_numeric and type2.is_numeric:
            return self.promote_numeric(type1, type2)
        elif type1.is_array and type2.is_array:
            return self.promote_arrays(type1, type2)
        elif type1 in string_types and type2 in string_types:
            return c_string_type
        elif type1.is_bool and type2.is_bool:
            return bool_
        else:
            raise minierror.UnpromotableTypeError((type1, type2))


def map_dtype(dtype):
    """
    Map a NumPy dtype to a minitype.

    >>> import numpy as np
    >>> map_dtype(np.dtype(np.int32))
    int32
    >>> map_dtype(np.dtype(np.int64))
    int64
    >>> map_dtype(np.dtype(np.object))
    PyObject *
    >>> map_dtype(np.dtype(np.float64))
    float64
    >>> map_dtype(np.dtype(np.complex128))
    complex128
    """
    import numpy as np

    if dtype.byteorder not in ('=', nbo, '|') and dtype.kind in ('iufbc'):
        raise minierror.UnmappableTypeError(
                "Only native byteorder is supported", dtype)

    item_idx = int(math.log(dtype.itemsize, 2))
    if dtype.kind == 'i':
        return [int8, int16, int32, int64][item_idx]
    elif dtype.kind == 'u':
        return [uint8, uint16, uint32, uint64][item_idx]
    elif dtype.kind == 'f':
        if dtype.itemsize == 2:
            pass # half floats not supported yet
        elif dtype.itemsize == 4:
            return float32
        elif dtype.itemsize == 8:
            return float64
        elif dtype.itemsize == 16:
            return float128
    elif dtype.kind == 'b':
        return int8
    elif dtype.kind == 'c':
        if dtype.itemsize == 8:
            return complex64
        elif dtype.itemsize == 16:
            return complex128
        elif dtype.itemsize == 32:
            return complex256
    elif dtype.kind == 'V':
        fields = [(name, map_dtype(dtype.fields[name][0]))
                      for name in dtype.names]
        is_aligned = dtype.alignment != 1
        return struct(fields, packed=not getattr(dtype, 'isalignedstruct',
                                                 is_aligned))
    elif dtype.kind == 'O':
        return object_

def create_dtypes():
    import numpy as np

    minitype2dtype = {
        int8     : np.int8,
        int16    : np.int16,
        int32    : np.int32,
        int64    : np.int64,
        uint8    : np.uint8,
        uint16   : np.uint16,
        uint32   : np.uint32,
        uint64   : np.uint64,

        float_   : np.float32,
        double   : np.float64,
        longdouble: np.longdouble,

        short    : np.dtype('h'),
        int_     : np.dtype('i'),
        long_    : np.dtype('l'),
        longlong : np.longlong,
        ushort   : np.dtype('H'),
        uint     : np.dtype('I'),
        ulong    : np.dtype('L'),
        ulonglong: np.ulonglong,

        complex64: np.complex64,
        complex128: np.complex128,
        complex256: getattr(np, 'complex256', None),

        bool_    : np.bool,
        object_  : np.object,
    }

    return dict((k, np.dtype(v)) for k, v in minitype2dtype.iteritems())

_dtypes = None
def map_minitype_to_dtype(type):
    global _dtypes

    if type.is_struct:
        import numpy as np

        fields = [(field_name, map_minitype_to_dtype(field_type))
                      for field_name, field_type in type.fields]
        return np.dtype(fields, align=not type.packed)

    if _dtypes is None:
        _dtypes = create_dtypes()

    if type.is_array:
        type = type.dtype

    dtype = _dtypes[type]
    assert dtype is not None, "dtype not supported in this numpy build"
    return dtype

def find_type_of_size(size, typelist):
    for type in typelist:
        if type.itemsize == size:
            return type

    assert False, "Type of size %d not found: %s" % (size, typelist)

NONE_KIND = 0
BOOL_KIND = 1
INT_KIND = 2
FLOAT_KIND = 3
COMPLEX_KIND = 4

class Type(miniutils.ComparableObjectMixin):
    """
    Base class for all types.

    .. attribute:: subtypes

        The list of subtypes to allow comparing and hashing them recursively
    """

    is_array = False
    is_pointer = False
    is_typewrapper = False

    is_bool = False
    is_numeric = False
    is_py_ssize_t = False
    is_char = False
    is_int = False
    is_float = False
    is_c_string = False
    is_object = False
    is_function = False
    is_int_like = False
    is_complex = False
    is_void = False

    kind = NONE_KIND

    subtypes = []

    mutated = True
    _ctypes_type = None

    def __init__(self, **kwds):
        vars(self).update(kwds)
        self.qualifiers = kwds.get('qualifiers', frozenset())

    def qualify(self, *qualifiers):
        "Qualify this type with a qualifier such as ``const`` or ``restrict``"
        qualifiers = list(qualifiers)
        qualifiers.extend(self.qualifiers)
        attribs = dict(vars(self), qualifiers=qualifiers)
        return type(self)(**attribs)

    def unqualify(self, *unqualifiers):
        "Remove the given qualifiers from the type"
        unqualifiers = set(unqualifiers)
        qualifiers = [q for q in self.qualifiers if q not in unqualifiers]
        attribs = dict(vars(self), qualifiers=qualifiers)
        return type(self)(**attribs)

    def pointer(self):
        "Get a pointer to this type"
        return PointerType(self)

    @property
    def subtype_list(self):
        return [getattr(self, subtype) for subtype in self.subtypes]

    @property
    def comparison_type_list(self):
        return self.subtype_list

    def _is_object(self, context):
        return context.is_object(self)

    def __eq__(self, other):
        # Don't use isinstance here, compare on exact type to be consistent
        # with __hash__. Override where sensible
        cmps = self.comparison_type_list
        if not cmps:
            return id(self) == id(other)

        return (isinstance(self, type(other)) and
                cmps == other.comparison_type_list)

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        cmps = self.comparison_type_list
        if not cmps:
            return hash(id(self))

        h = hash(type(self))
        for subtype in cmps:
            h = h ^ hash(subtype)

        return h

    def __getitem__(self, item):
        """
        Support array type creation by slicing, e.g. double[:, :] specifies
        a 2D strided array of doubles. The syntax is the same as for
        Cython memoryviews.
        """
        assert isinstance(item, (tuple, slice))

        def verify_slice(s):
            if s.start or s.stop or s.step not in (None, 1):
                raise minierror.InvalidTypeSpecification(
                    "Only a step of 1 may be provided to indicate C or "
                    "Fortran contiguity")

        if isinstance(item, tuple):
            step_idx = None
            for idx, s in enumerate(item):
                verify_slice(s)
                if s.step and (step_idx or idx not in (0, len(item) - 1)):
                    raise minierror.InvalidTypeSpecification(
                        "Step may only be provided once, and only in the "
                        "first or last dimension.")

                if s.step == 1:
                    step_idx = idx

            return ArrayType(self, len(item),
                             is_c_contig=step_idx == len(item) - 1,
                             is_f_contig=step_idx == 0)
        else:
            verify_slice(item)
            return ArrayType(self, 1, is_c_contig=bool(item.step))

    def declare(self):
        return str(self)

    def to_llvm(self, context):
        "Get a corresponding llvm type from this type"
        return context.to_llvm(self)

    def to_ctypes(self):
        """
        Convert type to ctypes. The result may be cached!
        """
        if self._ctypes_type is None:
            from . import ctypes_conversion
            self._ctypes_type = ctypes_conversion.convert_to_ctypes(self)
            self.mutated = False

        assert not self.mutated
        return self._ctypes_type

    def get_dtype(self):
        return map_minitype_to_dtype(self)

    def is_string(self):
        return self.is_c_string or self == char.pointer()

    def __getattr__(self, attr):
        if attr.startswith('is_'):
            return False
        return getattr(type(self), attr)

    def __call__(self, *args):
        """Return a FunctionType with return_type and args set
        """
        if len(args) == 1 and not isinstance(args[0], Type):
            # Cast in Python space
            # TODO: Create proxy object
            # TODO: Fully customizable type system (do this in Numba, not
            #       minivect)
            return args[0]

        return FunctionType(self, args)

class KeyHashingType(Type):

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other):
        return hasattr(other, 'key') and self.key == other.key

class ArrayType(Type):
    """
    An array type. ArrayType may be sliced to obtain a subtype:

    >>> double[:, :, ::1][1:]
    double[:, ::1]
    >>> double[:, :, ::1][:-1]
    double[:, :]

    >>> double[::1, :, :][:-1]
    double[::1, :]
    >>> double[::1, :, :][1:]
    double[:, :]
    """

    is_array = True
    subtypes = ['dtype']

    def __init__(self, dtype, ndim, is_c_contig=False, is_f_contig=False,
                 inner_contig=False, broadcasting=None):
        super(ArrayType, self).__init__()
        assert dtype is not None
        self.dtype = dtype
        self.ndim = ndim
        self.is_c_contig = is_c_contig
        self.is_f_contig = is_f_contig
        if ndim == 1 and (is_c_contig or is_f_contig):
            self.is_c_contig = True
            self.is_f_contig = True
        self.inner_contig = inner_contig or is_c_contig or is_f_contig
        self.broadcasting = broadcasting

    @property
    def comparison_type_list(self):
        return [self.dtype, self.is_c_contig, self.is_f_contig,
                self.inner_contig, self.ndim]

    def pointer(self):
        raise Exception("You probably want a pointer type to the dtype")

    def to_llvm(self, context):
        # raise Exception("Obtain a pointer to the dtype and convert that "
        #                 "to an LLVM type")
        return context.to_llvm(self)

    def __repr__(self):
        axes = [":"] * self.ndim
        if self.is_c_contig and self.ndim > 0:
            axes[-1] = "::1"
        elif self.is_f_contig and self.ndim > 0:
            axes[0] = "::1"

        return "%s[%s]" % (self.dtype, ", ".join(axes))

    def copy(self, **kwargs):
        if 'dtype' in kwargs:
            assert kwargs['dtype'] is not None
        array_type = copy.copy(self)
        vars(array_type).update(kwargs)
        return array_type

    @property
    def strided(self):
        type = self.copy()
        type.is_c_contig = False
        type.is_f_contig = False
        type.inner_contig = False
        type.broadcasting = None
        return type

    def __getitem__(self, index):
        assert isinstance(index, slice)
        assert index.step is None
        assert index.start is not None or index.stop is not None

        start = 0
        stop = self.ndim
        if index.start is not None:
            start = index.start
        if index.stop is not None:
            stop = index.stop

        ndim = len(range(self.ndim)[start:stop])

        if ndim == 0:
            type = self.dtype
        elif ndim > 0:
            type = self.strided
            type.ndim = ndim
            type.is_c_contig = self.is_c_contig and stop == self.ndim
            type.is_f_contig = self.is_f_contig and start == 0
            type.inner_contig = type.is_c_contig or type.is_f_contig
            if type.broadcasting:
                type.broadcasting = self.broadcasting[start:stop]
        else:
            raise IndexError(index, ndim)

        return type


class PointerType(Type):
    is_pointer = True
    subtypes = ['base_type']

    def __init__(self, base_type, **kwds):
        super(PointerType, self).__init__(**kwds)
        self.base_type = base_type
        self.itemsize = struct_.calcsize("P")

    def __repr__(self):
        return "%s *%s" % (self.base_type, " ".join(self.qualifiers))

    def to_llvm(self, context):
        if self.base_type.is_void:
            llvm_base_type = int_.to_llvm(context)
        else:
            llvm_base_type = self.base_type.to_llvm(context)

        return llvm.core.Type.pointer(llvm_base_type)

class CArrayType(Type):
    is_carray = True
    subtypes = ['base_type']

    def __init__(self, base_type, size, **kwds):
        super(CArrayType, self).__init__(**kwds)
        self.base_type = base_type
        self.size = size

    def __repr__(self):
        return "%s[%d]" % (self.base_type, self.size)

    def to_llvm(self, context):
        return llvm.core.Type.array(self.base_type.to_llvm(context), self.size)

class TypeWrapper(Type):
    is_typewrapper = True
    subtypes = ['opaque_type']

    def __init__(self, opaque_type, context, **kwds):
        super(TypeWrapper, self).__init__(**kwds)
        self.opaque_type = opaque_type
        self.context = context

    def __repr__(self):
        return self.context.declare_type(self)

    def __deepcopy__(self, memo):
        return self

class NamedType(Type):
    name = None

    def __eq__(self, other):
        return isinstance(other, NamedType) and self.name == other.name

    __hash__ = Type.__hash__ # !@#$ py3k

    def __repr__(self):
        if self.qualifiers:
            return "%s %s" % (self.name, " ".join(self.qualifiers))
        return str(self.name)

class NumericType(NamedType):
    """
    Base class for numeric types.

    .. attribute:: name

        name of the type

    .. attribute:: itemsize

        sizeof(type)

    .. attribute:: rank

        ordering of numeric types
    """
    is_numeric = True

class IntType(NumericType):
    is_int = True
    is_int_like = True
    name = "int"
    signed = True
    rank = 4
    itemsize = 4
    typecode = None

    kind = INT_KIND

    def __init__(self, typecode=None, **kwds):
        super(IntType, self).__init__(**kwds)
        self.typecode = typecode
        if typecode is not None:
            self.itemsize = struct_.calcsize(typecode)

    def to_llvm(self, context):
        if self.itemsize == 1:
            return lc.Type.int(8)
        elif self.itemsize == 2:
            return lc.Type.int(16)
        elif self.itemsize == 4:
            return lc.Type.int(32)
        else:
            assert self.itemsize == 8, self
            return lc.Type.int(64)

    def declare(self):
        if self.name.endswith(('16', '32', '64')):
            return self.name + "_t"
        else:
            return str(self)


class BoolType(IntType):
    is_bool = True
    name = "bool"
    kind = BOOL_KIND
    rank = 0

    def __repr__(self):
        return ("int %s" % " ".join(self.qualifiers)).rstrip()

    def __str__(self):
        return ("bool %s" % " ".join(self.qualifiers)).rstrip()

    def to_llvm(self, context):
        return llvm.core.Type.int(1)


class FloatType(NumericType):
    is_float = True

    kind = FLOAT_KIND

    def declare(self):
        if self.itemsize == 4:
            return "float"
        elif self.itemsize == 8:
            return "double"
        else:
            return str(self)

    @property
    def comparison_type_list(self):
        return self.subtype_list + [self.itemsize]

    def __eq__(self, other):
        return isinstance(other, FloatType) and self.itemsize == other.itemsize

    __hash__ = NumericType.__hash__

    def to_llvm(self, context):
        if self.itemsize == 4:
            return lc.Type.float()
        elif self.itemsize == 8:
            return lc.Type.double()
        else:
            is_ppc, is_x86 = get_target_triple()
            if self.itemsize == 16:
                if is_ppc:
                    return lc.Type.ppc_fp128()
                else:
                    return lc.Type.fp128()
            else:
                assert self.itemsize == 10 and is_x86
                return lc.Type.x86_fp80()

class ComplexType(NumericType):
    is_complex = True
    subtypes = ['base_type']

    kind = COMPLEX_KIND

class Py_ssize_t_Type(IntType):
    is_py_ssize_t = True
    name = "Py_ssize_t"
    rank = 9
    signed = True

    def __init__(self, **kwds):
        super(Py_ssize_t_Type, self).__init__(**kwds)
        if have_ctypes:
            if hasattr(ctypes, 'c_ssize_t'):
                self.itemsize = ctypes.sizeof(ctypes.c_ssize_t)
            else:
                self.itemsize = size_t.itemsize
        else:
            self.itemsize = _plat_bits // 8


class NPyIntp(IntType):
    is_numpy_intp = True
    name = "npy_intp"
    rank = 10

    def __init__(self, **kwds):
        super(NPyIntp, self).__init__(**kwds)
        ctypes_array = np.empty(0).ctypes.strides
        self.itemsize = ctypes.sizeof(ctypes_array._type_)

class CharType(IntType):
    is_char = True
    name = "char"
    rank = 1
    signed = True

    def to_llvm(self, context):
        return lc.Type.int(8)

class CStringType(Type):
    is_c_string = True

    def __repr__(self):
        return "const char *"

    def to_llvm(self, context):
        return char.pointer().to_llvm(context)

class VoidType(NamedType):
    is_void = True
    name = "void"

    def to_llvm(self, context):
        return lc.Type.void()

class ObjectType(Type):
    is_object = True
    itemsize = VoidType().pointer().itemsize

    def __repr__(self):
        return "PyObject *"

def pass_by_ref(type):
    return type.is_struct or type.is_complex

class Function(object):
    """
    Function types may be called with Python functions to create a Function
    object. This may be used to minivect users for their own purposes. e.g.

    @double(double, double)
    def myfunc(...):
       ...
    """
    def __init__(self, signature, py_func):
        self.signature = signature
        self.py_func = py_func

    def __call__(self, *args, **kwargs):
        """
        Implement this to pass the callable test for classmethod/staticmethod.
        E.g.

            @classmethod
            @void()
            def m(self):
                ...
        """
        raise minierror.Error("Not a callable function")

class FunctionType(Type):
    subtypes = ['return_type', 'args']
    is_function = True
    is_vararg = False

    struct_by_reference = False

    def __init__(self, return_type, args, name=None, is_vararg=False, **kwds):
        super(FunctionType, self).__init__(**kwds)
        self.return_type = return_type
        self.args = tuple(args)
        self.name = name
        self.is_vararg = is_vararg

    def to_llvm(self, context):
        assert self.return_type is not None
        self = self.actual_signature
        arg_types = [arg_type.pointer() if arg_type.is_function else arg_type
                         for arg_type in self.args]
        return lc.Type.function(self.return_type.to_llvm(context),
                                [arg_type.to_llvm(context)
                                     for arg_type in arg_types],
                                self.is_vararg)

    def __repr__(self):
        args = [str(arg) for arg in self.args]
        if self.is_vararg:
            args.append("...")
        if self.name:
            namestr = self.name
        else:
            namestr = ''
        return "%s (*%s)(%s)" % (self.return_type, namestr, ", ".join(args))

    @property
    def actual_signature(self):
        """
        Passing structs by value is not properly supported for different
        calling conventions in LLVM, so we take an extra argument
        pointing to a caller-allocated struct value.
        """
        if self.struct_by_reference:
            args = []
            for arg in self.args:
                if pass_by_ref(arg):
                    arg = arg.pointer()
                args.append(arg)

            return_type = self.return_type
            if pass_by_ref(self.return_type):
                return_type = void
                args.append(self.return_type.pointer())

            self = FunctionType(return_type, args)

        return self

    @property
    def struct_return_type(self):
        # Function returns a struct.
        return self.return_type.pointer()

    def __call__(self, *args):
        if len(args) != 1 or isinstance(args[0], Type):
            return super(FunctionType, self).__call__(*args)

        assert self.return_type is not None
        assert self.args is not None
        func, = args
        return Function(self, func)

class VectorType(Type):
    subtypes = ['element_type']
    is_vector = True
    vector_size = None

    def __init__(self, element_type, vector_size, **kwds):
        super(VectorType, self).__init__(**kwds)
        assert ((element_type.is_int or element_type.is_float) and
                element_type.itemsize in (4, 8)), element_type
        self.element_type = element_type
        self.vector_size = vector_size
        self.itemsize = element_type.itemsize * vector_size

    def to_llvm(self, context):
        return lc.Type.vector(self.element_type.to_llvm(context),
                              self.vector_size)

    @property
    def comparison_type_list(self):
        return self.subtype_list + [self.vector_size]

    def __repr__(self):
        itemsize = self.element_type.itemsize
        if self.element_type.is_float:
            if itemsize == 4:
                return '__m128'
            else:
                return '__m128d'
        else:
            if itemsize == 4:
                return '__m128i'
            else:
                raise NotImplementedError


def _sort_types_key(field_type):
    if field_type.is_complex:
        return field_type.base_type.rank * 2
    elif field_type.is_numeric or field_type.is_struct:
        return field_type.rank
    elif field_type.is_vector:
        return _sort_types_key(field_type.element_type) * field_type.vector_size
    elif field_type.is_carray:
        return _sort_types_key(field_type.base_type) * field_type.size
    elif field_type.is_pointer or field_type.is_object or field_type.is_array:
        return 8
    else:
        return 1

def _sort_key(keyvalue):
    field_name, field_type = keyvalue
    return _sort_types_key(field_type)

def sort_types(types_dict):
    # reverse sort on rank, forward sort on name
    d = {}
    for field in types_dict.iteritems():
        key = _sort_key(field)
        d.setdefault(key, []).append(field)

    def key(keyvalue):
        field_name, field_type = keyvalue
        return field_name

    fields = []
    for rank in sorted(d, reverse=True):
        fields.extend(sorted(d[rank], key=key))

    return fields

class struct(Type):
    """
    Create a struct type. Fields may be ordered or unordered. Unordered fields
    will be ordered from big types to small types (for better alignment).

    >>> struct([('a', int_), ('b', float_)], name='Foo') # ordered struct
    struct Foo { int a, float b }
    >>> struct(a=int_, b=float_, name='Foo') # unordered struct
    struct Foo { float b, int a }
    >>> struct(a=int32, b=int32, name='Foo') # unordered struct
    struct Foo { int32 a, int32 b }

    >>> S = struct(a=complex128, b=complex64, c=struct(f1=double, f2=double, f3=int32))
    >>> S
    struct { struct { double f1, double f2, int32 f3 } c, complex128 a, complex64 b }

    >>> S.offsetof('a')
    24
    """

    is_struct = True

    def __init__(self, fields=(), name=None, readonly=False, packed=False, **kwargs):
        super(struct, self).__init__()
        if fields and kwargs:
            raise minierror.InvalidTypeSpecification(
                    "The struct must be either ordered or unordered")

        if kwargs:
            fields = sort_types(kwargs)

        self.fields = list(fields)
        self.name = name
        self.readonly = readonly
        self.fielddict = dict(self.fields)
        self.packed = packed

        self.update_mutated()

    def copy(self):
        return struct(self.fields, self.name, self.readonly, self.packed)

    def __eq__(self, other):
        return other.is_struct and self.fields == other.fields

    def __repr__(self):
        if self.name:
            name = self.name + ' '
        else:
            name = ''
        return 'struct %s{ %s }' % (
                name, ", ".join("%s %s" % (field_type, field_name)
                                    for field_name, field_type in self.fields))

    def to_llvm(self, context):
        if self.packed:
            lstruct = llvm.core.Type.packed_struct
        else:
            lstruct = llvm.core.Type.struct

        return lstruct([field_type.to_llvm(context)
                           for field_name, field_type in self.fields])

    @property
    def comparison_type_list(self):
        return self.fields

    @property
    def subtype_list(self):
        return [field[1] for field in self.fields]

    def __hash__(self):
        return hash(tuple(self.fields))

    def is_prefix(self, other_struct):
        other_fields = other_struct.fields[:len(self.fields)]
        return self.fields == other_fields

    def add_field(self, name, type):
        assert name not in self.fielddict
        self.fielddict[name] = type
        self.fields.append((name, type))
        self.mutated = True

    def update_mutated(self):
        self.rank = sum([_sort_key(field) for field in self.fields])
        self.mutated = False

    def offsetof(self, field_name):
        """
        Compute the offset of a field. Must be used only after mutation has
        finished.
        """
        ctype = self.to_ctypes()
        return getattr(ctype, field_name).offset

def getsize(ctypes_name, default):
    try:
        return ctypes.sizeof(getattr(ctypes, ctypes_name))
    except ImportError:
        return default

def get_target_triple():
    target_machine = llvm.ee.TargetMachine.new()
    is_ppc = target_machine.triple.startswith("ppc")
    is_x86 = target_machine.triple.startswith("x86")
    return is_ppc, is_x86

#
### Internal types
#
c_string_type = CStringType()
void = VoidType()

#
### Public types
#
try:
    npy_intp = NPyIntp()
except ImportError:
    npy_intp = None

size_t = IntType(name="size_t", rank=8.5,
                 itemsize=getsize('c_size_t', _plat_bits // 8), signed=False)
Py_ssize_t = Py_ssize_t_Type()
Py_uintptr_t = IntType(name='Py_uintptr_t',
                       itemsize=getsize('c_void_p', Py_ssize_t.itemsize),
                       rank=8.5)

char = CharType(name="char", typecode='b')
short = IntType(name="short", rank=2, typecode='h')
int_ = IntType(name="int", rank=4, typecode='i')
long_ = IntType(name="long", rank=5, typecode='l')
longlong = IntType(name="PY_LONG_LONG", rank=8, typecode='q')

uchar = CharType(name="unsigned char", signed=False, typecode='B')
ushort = IntType(name="unsigned short", rank=2.5,
                 typecode='H', signed=False)
uint = IntType(name="unsigned int", rank=4.5, typecode='I', signed=False)
ulong = IntType(name="unsigned long", rank=5.5, typecode='L', signed=False)
ulonglong = IntType(name="unsigned PY_LONG_LONG", rank=8.5,
                    typecode='Q', signed=False)

float_ = FloatType(name="float", rank=20, itemsize=4)
double = FloatType(name="double", rank=21, itemsize=8)
longdouble = FloatType(name="long double", rank=22,
                       itemsize=ctypes.sizeof(ctypes.c_longdouble))

bool_ = BoolType()
object_ = ObjectType()

int8 = IntType(name="int8", rank=1, itemsize=1)
int16 = IntType(name="int16", rank=2, itemsize=2)
int32 = IntType(name="int32", rank=4, itemsize=4)
int64 = IntType(name="int64", rank=8, itemsize=8)

uint8 = IntType(name="uint8", rank=1.5, signed=False, itemsize=1)
uint16 = IntType(name="uint16", rank=2.5, signed=False, itemsize=2)
uint32 = IntType(name="uint32", rank=4.5, signed=False, itemsize=4)
uint64 = IntType(name="uint64", rank=8.5, signed=False, itemsize=8)

float32 = FloatType(name="float32", rank=20, itemsize=4)
float64 = FloatType(name="float64", rank=21, itemsize=8)
float128 = FloatType(name="float128", rank=22, itemsize=16)

complex64 = ComplexType(name="complex64", base_type=float32,
                        rank=30, itemsize=8)
complex128 = ComplexType(name="complex128", base_type=float64,
                         rank=31, itemsize=16)
complex256 = ComplexType(name="complex256", base_type=float128,
                         rank=32, itemsize=32)

integral = []
native_integral = []
floating = []
complextypes = []

for typename in __all__:
    minitype = globals()[typename]
    if minitype is None:
        continue

    if minitype.is_int:
        integral.append(minitype)
    elif minitype.is_float:
        floating.append(minitype)
    elif minitype.is_complex:
        complextypes.append(minitype)

numeric = integral + floating + complextypes
native_integral.extend((Py_ssize_t, size_t))

integral.sort(key=_sort_types_key)
native_integral = [minitype for minitype in integral
                                if minitype.typecode is not None]
floating.sort(key=_sort_types_key)
complextypes.sort(key=_sort_types_key)

def get_utility():
    import numpy

    return textwrap.dedent("""\
    #include <stdint.h>

    #ifndef HAVE_LONGDOUBLE
        #define HAVE_LONGDOUBLE %d
    #endif

    typedef struct {
        float real;
        float imag;
    } complex64;

    typedef struct {
        double real;
        double imag;
    } complex128;

    #if HAVE_LONGDOUBLE
    typedef struct {
        long double real;
        long double imag;
    } complex256;
    #endif

    typedef float float32;
    typedef double float64;
    #if HAVE_LONGDOUBLE
    typedef long double float128;
    #endif
    """ % hasattr(numpy, 'complex256'))

if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Promote and demote values of differing types in a minivect AST. This is run
before code generation. In LLVM types need to be equivalent for binary
operations.
"""
from __future__ import print_function, division, absolute_import

import sys
import copy

from . import minivisitor
from . import miniutils
from . import minitypes
from . import minierror

comparison_ops = set(['<', '>', '==', '!=', '>=', '<=',])

class TypePromoter(minivisitor.GenericTransform):
    """
    Promote and demote values of differing types.
    """

    def resolve_type(self, node):
        if node.type.is_array:
            node.type = node.type.dtype

    def promote(self, dst_type, node):
        return self.context.astbuilder.promote(dst_type, node)

    def visit_UnopNode(self, node):
        self.resolve_type(node)
        node.operand = self.promote(node.type, self.visit(node.operand))
        return node

    def visit_BinopNode(self, node):
        self.visitchildren(node)
        self.resolve_type(node)

        if node.operator in comparison_ops:
            dst_type = self.context.promote_types(node.lhs.type, node.rhs.type)
        else:
            dst_type = node.type

        if dst_type.is_pointer:
            return node

        return self.handle_binop(dst_type, node)

    def visit_VectorStoreNode(self, node):
        self.visitchildren(node)
        return node

    def handle_binop(self, dst_type, node):
        node.lhs = self.promote(dst_type, node.lhs)
        node.rhs = self.promote(dst_type, node.rhs)
        return node

    def visit_AssignmentExpr(self, node):
        self.visitchildren(node)
        self.resolve_type(node)
        return self.handle_binop(node.lhs.type, node)

    def visit_ResolvedVariable(self, node):
        return self.visit(node.element)



# -*- coding: utf-8 -*-
from __future__ import print_function, absolute_import
from llvm_cbuilder import *
from llvm_cbuilder import shortnames as C

from numba.utility.cbuilder.library import register
from numba.utility.cbuilder.numbacdef import NumbaCDefinition, from_numba

def get_constants(cbuilder):
    zero = cbuilder.constant(C.npy_intp, 0)
    one = cbuilder.constant(C.npy_intp, 1)
    return one, zero

# @register
class SliceArray(CDefinition):

    _name_ = "slice"
    _retty_ = C.char_p
    _argtys_ = [
        ('data', C.char_p),

        ('in_shape', C.pointer(C.npy_intp)),
        ('in_strides', C.pointer(C.npy_intp)),

        ('out_shape', C.pointer(C.npy_intp)),
        ('out_strides', C.pointer(C.npy_intp)),

        ('start', C.npy_intp),
        ('stop', C.npy_intp),
        ('step', C.npy_intp),

        ('src_dim', C.int),
        ('dst_dim', C.int),
    ]

    def _adjust_given_index(self, extent, negative_step, index, is_start):
        # Tranliterate the below code to llvm cbuilder

        # TODO: write in numba

        # For the start index in start:stop:step, do:
        # if have_start:
        #     if start < 0:
        #         start += shape
        #         if start < 0:
        #             start = 0
        #     elif start >= shape:
        #         if negative_step:
        #             start = shape - 1
        #         else:
        #             start = shape
        # else:
        #     if negative_step:
        #         start = shape - 1
        #     else:
        #         start = 0

        # For the stop index, do:
        # if stop is not None:
        #     if stop < 0:
        #         stop += extent
        #         if stop < 0:
        #             stop = 0
        #     elif stop > extent:
        #         stop = extent
        # else:
        #     if negative_step:
        #         stop = -1
        #     else:
        #         stop = extent

        one, zero = get_constants(self)

        with self.ifelse(index < zero) as ifelse:
            with ifelse.then():
                index += extent
                with self.ifelse(index < zero) as ifelse_inner:
                    with ifelse_inner.then():
                        index.assign(zero)

            with ifelse.otherwise():
                with self.ifelse(index >= extent) as ifelse:
                    with ifelse.then():
                        if is_start:
                            # index is 'start' index
                            with self.ifelse(negative_step) as ifelse:
                                with ifelse.then():
                                    index.assign(extent - one)
                                with ifelse.otherwise():
                                    index.assign(extent)
                        else:
                            # index is 'stop' index. Stop is exclusive, to
                            # we don't care about the sign of the step
                            index.assign(extent)

    def _set_default_index(self, default1, default2, negative_step, index):
        with self.ifelse(negative_step) as ifelse:
            with ifelse.then():
                index.assign(default1)
            with ifelse.otherwise():
                index.assign(default2)

    def adjust_index(self, extent, negative_step, index, default1, default2,
                     is_start=False, have_index=True):
        if have_index:
            self._adjust_given_index(extent, negative_step, index, is_start)
        else:
            self._set_default_index(default1, default2, negative_step, index)

    def body(self, data, in_shape, in_strides, out_shape, out_strides,
             start, stop, step, src_dim, dst_dim):

        stride = in_strides[src_dim]
        extent = in_shape[src_dim]

        one, zero = get_constants(self)
        if not self.have_step:
            step = one

        negative_step = step < zero

        self.adjust_index(extent, negative_step, start,
                                  default1=extent - one, default2=zero,
                                  is_start=True, have_index=self.have_start)
        self.adjust_index(extent, negative_step, stop,
                                 default1=-one, default2=extent,
                                 have_index=self.have_stop)

        # self.debug("extent", extent)
        # self.debug("negative_step", negative_step.cast(C.npy_intp))
        # self.debug("start/stop/step", start, stop, step)
        new_extent = self.var(C.npy_intp)
        new_extent.assign((stop - start) / step)
        with self.ifelse((stop - start) % step != zero) as ifelse:
            with ifelse.then():
                new_extent += one

        with self.ifelse(new_extent < zero) as ifelse:
            with ifelse.then():
                new_extent.assign(zero)

        result = self.var(data.type, name='result')
        result.assign(data[start * stride:])
        out_shape[dst_dim] = new_extent
        # self.debug("new_extent", new_extent)
        out_strides[dst_dim] = stride * step

        self.ret(result)

    def specialize(self, context, have_start, have_stop, have_step):
        self.context = context

        self.have_start = have_start
        self.have_stop = have_stop
        self.have_step = have_step

        self._name_ = "slice_%s_%s_%s" % (have_start, have_stop, have_step)

@register
class IndexAxis(NumbaCDefinition):

    _name_ = "index"
    _retty_ = C.char_p
    _argtys_ = [
        ('data', C.char_p),
        ('in_shape', C.pointer(C.npy_intp)),
        ('in_strides', C.pointer(C.npy_intp)),
        ('src_dim', C.npy_intp),
        ('index', C.npy_intp),
    ]

    def body(self, data, in_shape, in_strides, src_dim, index):
        result = self.var(data.type, name='result')
        # self.debug("indexing...", src_dim, "stride", in_strides[src_dim])
        result.assign(data[in_strides[src_dim] * index:])
        self.ret(result)

@register
class NewAxis(NumbaCDefinition):

    _name_ = "newaxis"
    _argtys_ = [
        ('out_shape', C.pointer(C.npy_intp)),
        ('out_strides', C.pointer(C.npy_intp)),
        ('dst_dim', C.int),
    ]

    def body(self, out_shape, out_strides, dst_dim):
        one, zero = get_constants(self)
        out_shape[dst_dim] = one
        out_strides[dst_dim] = zero
        # self.debug("newaxis in dimension:", dst_dim)
        self.ret()

# TODO: Transliterate the below to a numba function

@register
class Broadcast(NumbaCDefinition):
    """
    Transliteration of

        @cname('__pyx_memoryview_broadcast')
        cdef bint __pyx_broadcast(Py_ssize_t *dst_shape,
                                  Py_ssize_t *input_shape,
                                  Py_ssize_t *strides,
                                  int max_ndim, int ndim,
                                  bint *p_broadcast) nogil except -1:
            cdef Py_ssize_t i
            cdef int dim_offset = max_ndim - ndim

            for i in range(ndim):
                src_extent = input_shape[i]
                dst_extent = dst_shape[i + dim_offset]

                if src_extent == 1:
                    p_broadcast[0] = True
                    strides[i] = 0
                elif dst_extent == 1:
                    dst_shape[i + dim_offset] = src_extent
                elif src_extent != dst_extent:
                    __pyx_err_extents(i, dst_shape[i], input_shape[i])
    """

    _name_ = "__numba_util_broadcast"
    _argtys_ = [
        ('dst_shape', C.pointer(C.npy_intp)),
        ('src_shape', C.pointer(C.npy_intp)),
        ('src_strides', C.pointer(C.npy_intp)),
        ('max_ndim', C.int),
        ('ndim', C.int),
    ]
    _retty_ = C.int

    def body(self, dst_shape, src_shape, src_strides, max_ndim, ndim):
        dim_offset = max_ndim - ndim

        # self.debug("max_ndim", max_ndim, "ndim:", ndim)

        def constants(type):
            return self.constant(type, 0), self.constant(type, 1)

        zero, one = constants(C.npy_intp)
        zero_int, one_int = constants(C.int)

        with self.for_range(ndim) as (loop, i):
            src_extent = src_shape[i]
            dst_extent = dst_shape[i + dim_offset]

            with self.ifelse(src_extent == one) as ifelse:
                with ifelse.then():
                    # p_broadcast[0] = True
                    src_strides[i] = zero
                    # self.debug("set stride = 0")
                with ifelse.otherwise():
                    with self.ifelse(dst_extent == one) as ifelse:
                        with ifelse.then():
                            dst_shape[i + dim_offset] = src_extent

                        with ifelse.otherwise():
                            with self.ifelse(src_extent != dst_extent) as ifelse:
                                with ifelse.then():
                                    # Shape mismatch
                                    self.ret(zero_int)

        self.ret(one_int)



# Example from Travis Oliphant

import ctypes as ct
import numpy.random as nr
import os.path
from numpy.distutils.misc_util import get_shared_lib_extension

mtrand = ct.CDLL(nr.mtrand.__file__)

# Should we parse this from randomkit.h in the numpy directory? 
RK_STATE_LEN = len(nr.get_state()[1])

class rk_state(ct.Structure):
    _fields_ = [("key", ct.c_ulong * RK_STATE_LEN),
               ("pos", ct.c_int),
               ("has_gauss", ct.c_int),
               ("gauss", ct.c_double),
               ("has_binomial", ct.c_int),
               ("psave", ct.c_double),
               ("nsave", ct.c_long),
               ("r", ct.c_double),
               ("q", ct.c_double),
               ("fm", ct.c_double),
               ("m", ct.c_long),
               ("p1", ct.c_double),
               ("xm", ct.c_double),
               ("xl", ct.c_double),
               ("xr", ct.c_double),               
               ("c", ct.c_double),
               ("laml", ct.c_double),
               ("lamr", ct.c_double),
               ("p2", ct.c_double),
               ("p3", ct.c_double),
               ("p4", ct.c_double)]

try:
    rk_randomseed = mtrand.rk_randomseed
    rk_seed = mtrand.rk_seed
    rk_gamma = mtrand.rk_gamma
    rk_normal = mtrand.rk_normal
except AttributeError as e:
    raise ImportError(str(e))

rk_randomseed.argtypes = [ct.POINTER(rk_state)]

rk_seed.restype = None
rk_seed.argtypes = [ct.c_long, ct.POINTER(rk_state)]

state = rk_state()
state_p = ct.pointer(state)
state_vp = ct.cast(state_p, ct.c_void_p)


rk_gamma.restype = ct.c_double
rk_gamma.argtypes = [ct.POINTER(rk_state), ct.c_double, ct.c_double]

rk_normal.restype = ct.c_double
rk_normal.argtypes = [ct.POINTER(rk_state), ct.c_double, ct.c_double]

def init():
    if rk_randomseed(state_p) != 0:
        raise ValueError("Cannot initialize the random number generator.")

init()



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import numba as nb
from numba import *
from numba.containers import orderedcontainer

import numpy as np

INITIAL_BUFSIZE = 10
SHRINK = 1.5
GROW = 2

def notimplemented(msg):
    raise NotImplementedError("'%s' method of type 'typedlist'" % msg)

_list_cache = {}

#-----------------------------------------------------------------------
# Runtime Constructor
#-----------------------------------------------------------------------

def typedlist(item_type, iterable=None):
    """
    >>> typedlist(int_)
    []
    >>> tlist = typedlist(int_, range(10))
    >>> tlist
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    >>> tlist[5]
    5L

    >>> typedlist(float_, range(10))
    [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
    """
    typedlist_ctor = compile_typedlist(item_type)
    return typedlist_ctor(iterable)

#-----------------------------------------------------------------------
# Typedlist implementation
#-----------------------------------------------------------------------

def compile_typedlist(item_type, _list_cache=_list_cache):
    # item_type_t = typesystem.CastType(item_type)
    # dtype_t = typesystem.NumpyDtypeType(item_type)

    if item_type in _list_cache:
        return _list_cache[item_type]

    dtype = item_type.get_dtype()
    methods = orderedcontainer.container_methods(item_type, notimplemented)

    @nb.jit(warn=False)
    class typedlist(object):
        @void(object_)
        def __init__(self, iterable):
            self.size = 0

            # TODO: Use length hint of iterable for initial buffer size
            self.buf = np.empty(INITIAL_BUFSIZE, dtype=dtype)

            # TODO: implement 'is'/'is not'
            if iterable != None:
                self.extend(iterable)

        # TODO: Jit __getitem__/__setitem__ of numba extension types

        __getitem__ = methods['getitem']
        __setitem__ = methods['setitem']
        append = methods['append']
        extend = methods['extend']
        index = methods['index']
        count = methods['count']

        @item_type()
        def pop(self):
            # TODO: Optional argument 'index'
            size = self.size - 1
            item = self.buf[size]
            self.size = size

            if INITIAL_BUFSIZE < size < self.buf.shape[0] / 2:
                self.buf.resize(int(SHRINK * size), refcheck=False)

            return item

        @void(Py_ssize_t, item_type)
        def insert(self, index, value):
            notimplemented("insert")

        @void(item_type)
        def remove(self, value):
            notimplemented("remove")

        @void()
        def reverse(self):
            buf = self.buf
            size = self.size - 1
            for i in range(self.size / 2):
                tmp = buf[i]
                buf[i] = buf[size - i]
                buf[size - i] = tmp

        @void()
        def sort(self):
            # TODO: optional arguments cmp, key, reverse
            self.buf[:self.size].sort()

        @Py_ssize_t()
        def __len__(self):
            return self.size

        @nb.c_string_type()
        def __repr__(self):
            buf = ", ".join([str(self.buf[i]) for i in range(self.size)])
            return "[" + buf + "]"

    _list_cache[item_type] = typedlist
    return typedlist


if __name__ == "__main__":
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
# -*- encoding: UTF-8 -*-

from . import minivisitor
from . import miniutils
from . import minitypes
from . import specializers

def admissible(broadcasting_tuple, n_loops):
    """
    Check for admissibility. Indicates whether partial hoisting is the most
    efficient thing to perform. See also partially_hoistable()
    """
    if len(broadcasting_tuple) < n_loops:
        # In this this situation, we pad with leading broadcasting dimensions.
        # This means we have to hoist all the way
        return False

    # Filter leading False values
    i = 0
    for i, broadcasting in enumerate(broadcasting_tuple):
        if broadcasting:
            break

    # Check for all trailing values (at least one) being True
    return broadcasting_tuple[i:] and miniutils.all(broadcasting_tuple[i:])

def partially_hoistable(broadcasting_tuple, n_loops):
    """
    This function indicates, when admissible() returns false, whether an
    expression is partially hoistable. This means the caller must establish
    whether repeated computation or an array temporary will be more beneficial.

    If the expression is a variable, there is no repeaetd computation, and
    it should be hoisted as far as possible.
    """
    return broadcasting_tuple[-1]


def broadcasting(broadcasting_tuple1, broadcasting_tuple2):
    return broadcasting_tuple1 != broadcasting_tuple2

class HoistBroadcastingExpressions(specializers.BaseSpecializer):
    """
    This transform hoists out part of sub-expressions which are broadcasting.
    There are two cases:

        1) We can hoist out the sub-expression and store it in a scalar for
           broadcasting
        2) We have to hoist the sub-expression out entirely and store it in
           a temporary array

    As an alternative to 2), we could swap axes to return to situation 1),
    i.e. reorder the element-wise traversal order. We do not choose this
    option, since the loop order is tailored to cache-friendliness.

    We determine where to hoist sub-expression to based on broadcasting
    information. Each type has a broadcasting tuple with a true/false value
    for each dimension, specifying whether it will broadcast in that dimension
    (i.e. broadcasting is not optional in that dimension).

    We make the following observations:

        1) Trailing truth values mean we can hoist the sub-expression out just
           before the first truth value in the consecutive sequence of truth
           values

            Example ``(False, True, True)``::

                A[:, None, None] * A[:, None, None] * B[:, :, :]

            becomes::

                for i in shape[0]:
                    temp = A[i, 0, 0] * A[i, 0, 0]
                    for j in shape[1]:
                        for k in shape[2]:
                            temp * B[i, j, k]

        2) If not all consecutive leading values are false, we have to assign
           to a temporary array (i.e., hoist out all the way)

            Example ``(True, True, False)``::

                A[None, None, :] * A[None, None, :] * B[:, :, :]

            becomes::

                allocate temp

                for k in shape[2]:
                    temp[k] = A[0, 0, k] * A[0, 0, k]

                for i in shape[0]:
                    for j in shape[1]:
                        for k in shape[2]:
                            temp[k] * B[i, j, k]

                deallocate temp

    More generally, if the index sequence of array A is not an admissible prefix
    of the total index sequence, we have situation 2). For instance,
    ``(True, False, True)`` would mean we could hoist out the expression one
    level, but we would still have repeated computation. What we could do in
    this case, in addition to 2), is reduce indexing overhead, i.e. generate::

        for j in shape[1]:
            temp[j] = A[0, j, 0] * A[0, j, 0]

        for i in shape[0]:
            for j in shape[1]:
                temp_scalar = temp[j]
                for k in shape[2]:
                    temp_scalar * B[i, j, k]

    This is bonus points.
    """

    def visit_FunctionNode(self, node):
        self.function = node
        inner_loop = node.for_loops[-1]
        self.visitchildren(inner_loop)
        return node

    def visit_Variable(self, node):
        type = node.type
        if type.is_array and type.broadcasting is not None:
            n_loops = len(self.function.for_loops)
            if admissible(type.broadcasting, n_loops):
                node.hoistable = True
            elif partially_hoistable(type.broadcasting, n_loops):
                # TODO: see whether `node` should be fully (array temporary)
                # TODO: or partially hoisted
                node.hoistable = True
            elif miniutils.any(type.broadcasting):
                pass # enable when temporaries are implemented in minivect
                # node.need_temp = True

            node.broadcasting = type.broadcasting

        return node

    def visit_ArrayAtribute(self, node):
        return node

    def _hoist_binop_operands(self, b, node):
        # perform hoisitng. Does not handle all expressions correctly yet.
        if not node.lhs.hoistable or not node.rhs.hoistable:
            if node.lhs.hoistable:
                node.lhs = self.hoist(node.lhs)
            else:
                node.rhs = self.hoist(node.rhs)

            return node

        lhs_hoisting_level = self.hoisting_level(node.lhs)
        rhs_hoisting_level = self.hoisting_level(node.rhs)

        if lhs_hoisting_level == rhs_hoisting_level:
            node.hoistable = True
            node.broadcasting = node.lhs.broadcasting
            return node

        def binop():
            result = b.binop(node.type, node.operator, node.lhs, node.rhs)
            result.broadcasting = broadcasting
            result.hoistable = True
            return result

        if lhs_hoisting_level < rhs_hoisting_level:
            broadcasting = node.rhs.broadcasting
            node.lhs = self.hoist(node.lhs)
            return self.hoist(binop())
        else: # lhs_hoisting_level > rhs_hoisting_level
            broadcasting = node.lhs.broadcasting
            node.rhs = self.hoist(node.rhs)
            return self.hoist(binop())

    def _make_temp_binop_operands(self, node):
        if broadcasting(node.lhs.broadcasting, node.rhs.broadcasting):
            node.need_temp = True
        else:
            if node.lhs.need_temp:
                node.lhs = self.make_temp(node.lhs)
            if node.rhs.need_temp:
                node.rhs = self.make_temp(node.rhs)

    def visit_BinaryOperationNode(self, node):
        b = self.astbuilder

        self.visitchildren(node)

        node.broadcasting = None

        if node.lhs.need_temp or node.rhs.need_temp:
            return self._make_temp_binop_operands(node)
        elif node.lhs.hoistable or node.rhs.hoistable:
            return self._hoist_binop_operands(b, node)

        return node

    def visit_ForNode(self, node):
        self.visitchildren(node)
        self.handle_pending_stats(node)
        return node

    def visit_BinopNode(self, node):
        # used in superclass, override here
        return self.visit_BinaryOperationNode(node)

    def visit_AssignmentExpr(self, node):
        rhs = self.visit(node.rhs)
        node.rhs = self.process_expr(rhs)
        return node

    def visit_UnopNode(self, node):
        o = node.operand = self.visit(node.operand)

        node.hoistable = o.hoistable
        node.need_temp = o.need_temp
        node.broadcasting = o.broadcasting

        return node

    def process_expr(self, expr):
        if expr.hoistable:
            return self.hoist(expr)
        elif expr.need_temp:
            return self.make_temp(expr)
        else:
            return expr

    def make_temp(self, node):
        "Not implemented yet"
        return node

    def hoisting_level(self, node):
        i = 0
        for i, broadcasting in enumerate(node.broadcasting[::-1]):
            if not broadcasting:
                break

        return self.function.ndim - 1 - i

    def hoist(self, node):
        if not node.hoistable:
            return node

        b = self.astbuilder

        hoisting_level = self.hoisting_level(node)
        if hoisting_level < 0:
            for_loop = self.function
        else:
            for_loop = self.function.for_loops[hoisting_level]

        temp = b.temp(node.type.dtype, name='hoisted_temp')
        temp.broadcasting = None

        # TODO: keep track of the variables
        for variable in self.treepath(node, '//Variable'):
            variable.hoisted = True

        stat = b.assign(temp, node, may_reorder=False)
        for_loop.body = b.stats(stat, for_loop.body)
        return self.visit(temp)



#! /usr/bin/env python
# ______________________________________________________________________
'''test_sum

Test the sum2d() example.
'''
# ______________________________________________________________________

import numpy

from numba import *
from numba.decorators import jit
from numba.testing import test_support

import sys
import unittest

# ______________________________________________________________________

def sum2d(arr):
    M, N = arr.shape
    result = 0.0
    for i in range(M):
        for j in range(N):
            result += arr[i,j]
    return result

# ______________________________________________________________________

def bad_sum2d(arr):
    '''Unit test code for issue #34:
    https://github.com/numba/numba/issues/34'''
    M, N = arr.shape
    result = 0.0
    for i in range(M):
        for j in range(N):
            result += arr[i,j]
        return result

# ______________________________________________________________________

class TestASTSum2d(test_support.ASTTestCase):
    
    def test_vectorized_sum2d(self):
        usum2d = self.jit(argtypes=[double[:,:]],
                          restype=double)(sum2d)
        image = numpy.random.rand(10, 10)
        plain_old_result = sum2d(image)
        hot_new_result = usum2d(image)
        self.assertTrue((abs(plain_old_result - hot_new_result) < 1e-9).all())

    def test_vectorized_sum2d(self):
        pass

    def test_vectorized_sum2d(self):
        usum2d = self.jit(argtypes=[double[:,:]],
                          restype=double)(sum2d)
        image = numpy.random.rand(10, 10)
        plain_old_result = sum2d(image)
        hot_new_result = usum2d(image)
        self.assertTrue((abs(plain_old_result - hot_new_result) < 1e-9).all())

    def _bad_sum2d(self):
        compiled_bad_sum2d = self.jit(argtypes = [double[:,:]],
                                      restype = double)(bad_sum2d)
        image = numpy.random.rand(10, 10)
        self.assertEqual(bad_sum2d(image), compiled_bad_sum2d(image))

    @test_support.checkSkipFlag("Test fails due to problem in Meta.")
    def test_bad_sum2d(self):
        self._bad_sum2d()

if __name__ == "__main__":
    #TestASTSum2d('test_vectorized_sum2d').debug()
    unittest.main()

# ______________________________________________________________________
# End of test_sum.py



# -*- coding: utf-8 -*-
"""
Types for closures and inner functions.
"""
from __future__ import print_function, division, absolute_import

from numba.minivect import minitypes
from numba.typesystem.basetypes import NumbaType
from numba.exttypes.types.extensiontype import ExtensionType

class ClosureType(NumbaType, minitypes.ObjectType):
    """
    Type of closures and inner functions.
    """

    is_closure = True

    def __init__(self, signature, **kwds):
        super(ClosureType, self).__init__(**kwds)
        self.signature = signature
        self.closure = None

    def add_scope_arg(self, scope_type):
        self.signature.args = (scope_type,) + self.signature.args

    def __repr__(self):
        return "<closure(%s)>" % self.signature

class ClosureScopeType(ExtensionType):
    """
    Type of the enclosing scope for closures. This is always passed in as
    first argument to the function.
    """

    is_closure_scope = True
    is_final = True

    def __init__(self, py_class, parent_scope, **kwds):
        super(ClosureScopeType, self).__init__(py_class, **kwds)
        self.parent_scope = parent_scope
        self.unmangled_symtab = None

        if self.parent_scope is None:
            self.scope_prefix = ""
        else:
            self.scope_prefix = self.parent_scope.scope_prefix + "0"

    def __repr__(self):
        return "closure_scope(%s)" % self.attribute_table



import ctypes
import numba
from numba import *
from numba.testing.test_support import testmod, autojit_py3doc

import numpy as np

int32p = int32.pointer()
voidp = void.pointer()

@autojit_py3doc
def test_pointer_arithmetic():
    """
    >>> test_pointer_arithmetic()
    48L
    """
    p = int32p(Py_uintptr_t(0))
    p = p + 10
    p += 2
    return Py_uintptr_t(p) # 0 + 4 * 12

@autojit_py3doc(locals={"pointer_value": Py_uintptr_t})
def test_pointer_indexing(pointer_value, type_p):
    """
    >>> a = np.array([1, 2, 3, 4], dtype=np.float32)
    >>> test_pointer_indexing(a.ctypes.data, float32.pointer())
    (1.0, 2.0, 3.0, 4.0)

    >>> a = np.array([1, 2, 3, 4], dtype=np.int64)
    >>> test_pointer_indexing(a.ctypes.data, int64.pointer())
    (1L, 2L, 3L, 4L)
    """
    p = type_p(pointer_value)
    return p[0], p[1], p[2], p[3]

if __name__ == '__main__':
#    print test_pointer_arithmetic()
#    a = np.array([1, 2, 3, 4], dtype=np.float32)
#    print test_pointer_indexing(a.ctypes.data, float32.pointer())
    pass

testmod()



"""
>>> native_convert(char, -10)
-10L
>>> native_convert(char, 10)
10L
>>> native_convert(char, 127)
127L

This doesn't work yet, we should get an error here. We don't get one because
autojit detects the int type which is natively truncated to a char.

TODO::::::::::

    >> native_convert(char, 128)

        => need exception!

>>> object_convert(char, 128)
Traceback (most recent call last):
    ...
OverflowError: value too large to convert to signed char
>>> object_convert(char, -128)
-128L
>>> object_convert(char, -129)
Traceback (most recent call last):
    ...
OverflowError: value too large to convert to signed char
>>> object_convert(char, 2.9)
2L

TODO:::::::::::

    Test all numeric types for overflows!

TODO:::::::::::

    Test typedef types (npy_intp, Py_uintptr_t, etc)
"""

from numba import *

@autojit
def native_convert(dst_type, value):
    return dst_type(value)

@autojit(locals=dict(obj=object_))
def object_convert(dst_type, obj):
    return dst_type(obj)

if __name__ == "__main__":
    import numba
    from numba.testing.test_support import rewrite_doc
    __doc__ = rewrite_doc(__doc__)
    numba.testing.testmod()



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import numba
from numba import *
from numba import error, nodes
from numba.type_inference import module_type_inference
from numba import typesystem

if PY3:
    import builtins
else:
    import __builtin__ as builtins

debug = False
#debug = True

def resolve_function(func_variable):
    "Get a function object given a function name"
    func = None
    func_type = func_variable.type

    if func_type.is_builtin:
        func = getattr(builtins, func_variable.name)
    elif func_type.is_global:
        func = func_type.value
    elif func_type.is_module_attribute:
        func = getattr(func_type.module, func_type.attr)
    elif func_type.is_autojit_function:
        func = func_type.autojit_func
    elif func_type.is_jit_function:
        func = func_type.jit_func

    return func

def infer_typefunc(context, call_node, func_type, default_node):
    func_var = call_node.func.variable
    if func_var.is_constant:
        func_type = typesystem.KnownValueType(func_var.constant_value)

    if (func_type.is_known_value and
            module_type_inference.is_registered(func_type.value)):
        # Try the module type inferers
        result_node = module_type_inference.resolve_call_or_none(
                                    context, call_node, func_type)
        if result_node:
            return result_node

    return default_node

def parse_signature(node, func_type):
    types = []
    for arg in node.args:
        if not arg.variable.type.is_cast:
            raise error.NumbaError(arg, "Expected a numba type")
        else:
            types.append(arg.variable.type)

    signature = func_type.dst_type(*types)
    new_node = nodes.const(signature, typesystem.CastType(signature))
    return new_node



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import sys
import numba

# TODO: Use argparse
if '--loop' in sys.argv:
    whitelist = [arg for arg in sys.argv[1:] if arg != '--loop']
    while True:
        exit_status = numba.test(whitelist)
        if exit_status != 0:
            sys.exit(exit_status)
else:
    sys.exit(numba.test(sys.argv[1:]))



from pprint import pprint

import numpy as np

import numba
from numba import *

from numba.testing.test_support import *
from numba.control_flow.tests.test_cfg_type_infer import infer as _infer
from numba import typesystem

T = numba.template()

@autojit_py3doc(T(T[:, :]), warn=False, locals=dict(scalar=T))
def test_simple_template(array):
    """
    >>> test_simple_template(np.arange(10, 12, dtype=np.float32))
    Traceback (most recent call last):
        ...
    InvalidTemplateError: Template expects 2 dimensions, got 1

    >>> test_simple_template(np.arange(10, 12, dtype=np.float32).reshape(2, 1))
    10.0

    #------------------------------------------------------------------------
    # Test type resolving
    #------------------------------------------------------------------------
    >>> infer(test_simple_template.py_func, double(double[:, :]), T(T[:, :]),
    ...       locals=dict(scalar=T))
    [('array', double[:, :]), ('scalar', double)]

    >>> infer(test_simple_template.py_func, double(double[:, :]), T(T[:, :]),
    ...       locals=dict(scalar=T.pointer()))
    Traceback (most recent call last):
        ...
    UnpromotableTypeError: (double *, double)

    #------------------------------------------------------------------------
    # Test type attributes
    #------------------------------------------------------------------------
    >>> infer(test_simple_template.py_func, double(double[:, :]), T.dtype(T),
    ...       locals=dict(scalar=T.dtype))
    [('array', double[:, :]), ('scalar', double)]
    """
    scalar = array[0, 0]
    return scalar

#------------------------------------------------------------------------
# Test type matching
#------------------------------------------------------------------------

T1 = numba.template("T1")
T2 = numba.template("T2")
T3 = numba.template("T3")
T4 = numba.template("T4")

A = T1[:, :]
F = void(T1)
S = numba.struct(a=T1, b=T2.pointer(), c=T3[:], d=void(T4))
P = T2.pointer()

type_context1 = { T1: int_, T2: float_, T3: double, T4: short, }
type_context2 = { T1: int_[:, :], T2: void(float_),
                  T3: numba.struct(a=double, b=float_), T4: short.pointer(), }

def test_type_matching(array, func, struct, pointer):
    """
    >>> infer(test_type_matching, template_signature=void(A, F, S, P),
    ...       type_context=type_context1)
    [('array', int[:, :]), ('func', void (*)(int)), ('pointer', float *), ('struct', struct { float * b, double[:] c, int a, void (*)(short) d })]
    """
    func(array[0, 0])
    struct.b = pointer

def test_type_attributes(array, func, struct, pointer):
    """
    >>> locals = dict(dtype=T1.dtype, arg=T2.args[0], field_a=T3.fielddict['a'],
    ...               field_b=T3.fielddict['b'], scalar=T4.base_type)
    >>> pprint(infer(test_type_attributes, template_signature=void(T1, T2, T3, T4),
    ...              type_context=type_context2, locals=locals))
    [('array', int[:, :]),
     ('func', void (*)(float)),
     ('pointer', short *),
     ('struct', struct { double a, float b }),
     ('arg', float),
     ('dtype', int),
     ('field_a', double),
     ('field_b', float),
     ('scalar', short)]
    """
    dtype = array[0, 0]
    arg = 0
    field_a = 0
    field_b = 0
    scalar = 0

@autojit_py3doc(T(T, double), locals=None)
def test_template_with_concretes(a, b):
    """
    >>> test_template_with_concretes(1, 2)
    3L
    """
    return a + b

@autojit(complex128(T, double), locals=None)
def test_template_with_concretes2(a, b):
    """
    >>> test_template_with_concretes2(1, 2)
    (3+0j)
    >>> test_template_with_concretes2(1.0, 2.0)
    (3+0j)
    >>> test_template_with_concretes2(1+0j, 2)
    (3+0j)

    >>> test_template_with_concretes2(1+0j, 2+0j)
    Traceback (most recent call last):
        ...
    TypeError: can't convert complex to float
    """
    return a + b


@autojit_py3doc(T2(T1, double), locals=None)
def test_unknown_template_error(a, b):
    """
    >>> test_unknown_template_error(1, 2)
    Traceback (most recent call last):
        ...
    InvalidTemplateError: Unknown template type: T2
    """
    return a + b

@autojit_py3doc(T(T, T), locals=None)
def test_template_inconsistent_types_error(a, b):
    """
    >>> test_template_inconsistent_types_error(1, 2)
    3L
    >>> test_template_inconsistent_types_error(1, 2.0)
    Traceback (most recent call last):
        ...
    InvalidTemplateError: Inconsistent types found for template: int and double
    """
    return a + b

#------------------------------------------------------------------------
# Test utilities
#------------------------------------------------------------------------

def infer(func, signature=None, template_signature=None,
          locals=None, type_context=None):

    if signature is None:
        signature = specialize(template_signature, type_context)

    if locals is not None:
        locals = dict(locals)

    sig, symbols = _infer(func, signature,
                          template_signature=template_signature,
                          locals=locals, warn=False)

    if locals is not None:
        local_vars = sorted(locals.iteritems())
    else:
        local_vars = []

    vars = sorted((name, var.type) for name, var in symbols.iteritems())
    return vars + local_vars

def specialize(T, context):
    return typesystem.resolve_template_type(T, context)


if __name__ == '__main__':
    testmod()



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
from numba.typesystem import *
from numba.minivect import minitypes

#------------------------------------------------------------------------
# Container Types
#------------------------------------------------------------------------

class ContainerListType(NumbaKeyHashingType, minitypes.ObjectType):

    is_container = True
    subtypes = ['base_type']

    def __init__(self, base_type, size=-1):
        super(ContainerListType, self).__init__()
        self.base_type = base_type
        self.size = size

    @property
    def key(self):
        return self.base_type, self.size

    def is_sized(self):
        return self.size >= 0

    def __repr__(self):
        return "%s(%s, %s)" % (self.name, self.base_type, self.size)

class TupleType(ContainerListType):

    is_tuple = True
    name = "tuple"

class ListType(ContainerListType):
    is_list = True
    name = "list"

class MapContainerType(NumbaType):

    is_map = True

    def __init__(self, key_type, value_type, size=-1):
        super(MapContainerType, self).__init__()
        self.key_type = key_type
        self.value_type = value_type
        self.size = size

class DictType(MapContainerType, minitypes.ObjectType):

    is_dict = True
    name = "dict"
    size = 0

    def __str__(self):
        return "dict(%s)" % ", ".join(["..."] * self.size)

#------------------------------------------------------------------------
# Natively Typed Container Types
#------------------------------------------------------------------------

class TypedContainerListType(ContainerListType):

    is_typed_container = True

class TypedMapContainerType(MapContainerType):

    is_typed_map = True

class TypedTupleType(TypedContainerListType):

    is_typed_tuple = True
    name = "typedtuple"

class TypedListType(TypedContainerListType):

    is_typed_list = True
    name = "typedlist"

class TypedSetType(TypedContainerListType):

    is_typed_set = True
    name = "typedset"

class TypedFrozenSetType(TypedContainerListType):

    is_typed_frozenset = True
    name = "typedfrozenset"

class TypedDictType(TypedMapContainerType, minitypes.ObjectType):

    is_typed_dict = True
    name = "typeddict"

#------------------------------------------------------------------------
# User-facing constructors
#------------------------------------------------------------------------

def typedlist(base_type, size=-1):
    return TypedTupleType(base_type, size)

def typedtuple(base_type, size=-1):
    return TypedListType(base_type, size)

def typedset(base_type, size=-1):
    return TypedSetType(base_type, size)

def typedfrozenset(base_type, size=-1):
    return TypedSetType(base_type, size)

def typeddict(base_type, size=-1):
    return TypedFrozenSetType(base_type, size)


#------------------------------------------------------------------------
# Shorthands
#------------------------------------------------------------------------

tuple_ = TupleType(object_)
list_ = ListType(object_)
dict_ = DictType(object_, object_)


from numba import *
from numba import error

#@autojit
def func():
    if x:
        print("hello")
    else:
        print("world")

def compile_func1():
    try:
        jit(void())(func)
    except error.NumbaError as e:
        print("exception: %s" % e)

__doc__ = """
>>> compile_func1()
--------------------- Numba Encountered Errors or Warnings ---------------------
    if x:
-------^
Error 6:7: No global named 'x'
--------------------------------------------------------------------------------
exception: 6:7: No global named 'x'
"""

#@autojit
def func2():
    print(10[20])

def compile_func2():
    try:
        jit(void())(func2)
    except error.NumbaError as e:
        print("exception: %s" % e)

__doc__ += """>>> compile_func2()
--------------------- Numba Encountered Errors or Warnings ---------------------
    print(10[20])
----------^
Error 29:10: object of type int cannot be indexed
--------------------------------------------------------------------------------
exception: 29:10: object of type int cannot be indexed
"""

@autojit # this often messes up line numbers
def func_decorated():
    print(10[20])

def compile_func3():
    try:
        func_decorated()
    except error.NumbaError as e:
        print("exception: %s" % e)

__doc__ += """
>>> compile_func3()
--------------------- Numba Encountered Errors or Warnings ---------------------
    print(10[20])
----------^
Error 48:10: object of type int cannot be indexed
--------------------------------------------------------------------------------
exception: 48:10: object of type int cannot be indexed
"""

if __name__ == '__main__':
    import numba
    numba.testing.testmod()



#! /usr/bin/env python
# ______________________________________________________________________
'''test_forloop

Test the Numba compiler on a simple for loop over an iterable object.
'''
# ______________________________________________________________________

import numba
from numba import *
from numba.decorators import autojit

from numba.minivect import minitypes
hash(minitypes.double[:])

import numpy
import numpy as np

import unittest

# ______________________________________________________________________

def _for_loop_fn_0 (iterable):
    acc = 0.
    for value in iterable:
        acc += value
    return acc

# ______________________________________________________________________

def _for_loop_fn_1 (start, stop, inc):
    acc = 0
    for value in range(start, stop, inc):
        acc += value
    return acc

# ______________________________________________________________________

def _for_loop_fn_2 (stop):
    acc = 0
    for value_0 in range(stop):
        for value_1 in range(stop):
            acc += value_0 * value_1
    return acc

# ______________________________________________________________________

def _for_loop_fn_3 (stop):
    acc = 0
    for i in range(stop):
        for j in range(stop):
            for k in range(stop):
                for l in range(stop):
                    acc += 1
    return acc

for_loop_fn_0 = autojit(backend='ast')(_for_loop_fn_0)
for_loop_fn_1 = autojit(backend='ast')(_for_loop_fn_1)
for_loop_fn_2 = autojit(backend='ast')(_for_loop_fn_2)
for_loop_fn_3 = autojit(backend='ast')(_for_loop_fn_3)

# ______________________________________________________________________

class TestForLoop(unittest.TestCase):
#    @unittest.skipUnless(__debug__, "Requires implementation of iteration "
#                         "over arrays.")
    def test_compiled_for_loop_fn_0(self):
        for dtype in (np.float32, np.float64, np.int32, np.int64):
            test_data = np.arange(10, dtype=dtype)
            result = for_loop_fn_0(test_data)
            self.assertEqual(result, 45)
            self.assertEqual(result, _for_loop_fn_0(test_data))

    def test_compiled_for_loop_fn_1(self):
        result = for_loop_fn_1(1, 4, 1)
        self.assertEqual(result, 6)
        self.assertEqual(result, _for_loop_fn_1(1, 4, 1))

    def test_compiled_for_loop_fn_2(self):
        result = for_loop_fn_2(4)
        self.assertEqual(result, 36)
        self.assertEqual(result, _for_loop_fn_2(4))

    def test_compiled_for_loop_fn_3(self):
        result = for_loop_fn_3(3)
        self.assertEqual(result, _for_loop_fn_3(3))
        self.assertEqual(result, 81)

# ______________________________________________________________________

if __name__ == "__main__":
    print((for_loop_fn_2(10)))
#    print for_loop_fn_2(10.0)
    unittest.main()

# ______________________________________________________________________
# End of test_forloop.py



"""
Test class attributes.
"""

import numba
from numba import *
from numba.testing.test_support import parametrize, main

def make_base(compiler):
    @compiler
    class Base(object):

        value1 = double
        value2 = int_

        @void(int_, double)
        def __init__(self, value1, value2):
            self.value1 = value1
            self.value2 = value2

        @void(int_)
        def setvalue(self, value):
            self.value1 = value

        @double()
        def getvalue1(self):
            return self.value1

    return Base

def make_derived(compiler):
    Base = make_base(compiler)

    @compiler
    class Derived(Base):

        value3 = float_

        @void(int_)
        def setvalue(self, value):
            self.value3 = value

    return Base, Derived

#------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------

@parametrize(jit, autojit)
def test_baseclass_attrs(compiler):
    Base = make_base(compiler)

    assert Base(10, 11.0).value1 == 10.0
    assert Base(10, 11.0).value2 == 11

    obj = Base(10, 11.0)
    obj.setvalue(12)
    assert obj.getvalue1() == 12.0

@parametrize(jit) #, autojit)
def test_derivedclass_attrs(compiler):
    Base, Derived = make_derived(compiler)

    obj = Derived(10, 11.0)
    obj.setvalue(9)
    assert obj.value3 == 9.0


if __name__ == '__main__':
    # test_derivedclass_attrs(autojit)
    main()



from numba import *
import numpy as np

def get_functions():
    def sqrt(a):
        return 2.7 + np.sqrt(a) + 1.6

    def log(a):
        return 2.7 + np.log(a) + 1.6

    def log10(a):
        return 2.7 + np.log10(a) + 1.6

    def log1p(a):
        return 2.7 + np.log1p(a) + 1.6

    def log2(a):
        return 2.7 + np.log2(a) + 1.6

    def exp(a):
        return 2.7 + np.exp(a) + 1.6

#    def expm1(a):
#        return 2.7 + np.expm1(a) + 1.6

    def sin(a):
        return 2.7 + np.sin(a) + 1.6

    def cos(a):
        return 2.7 + np.cos(a) + 1.6

    def absolute(a):
        return 2.7 + np.abs(a) + 1.6

    return locals()

dtypes = ['i', 'l', 'f', 'd', np.complex128]

def test_math_funcs():
    functions = get_functions()
    exceptions = 0
    for func_name in functions:
        # func_name = 'sqrt'
        func = functions[func_name]
        for dtype in dtypes:
            numba_func = autojit(func)

            x = np.arange(8 * 12, dtype=dtype).reshape(8, 12)
            x = ((x + 10) / 5).astype(dtype)

            r1 = numba_func(x)
            r2 = numba_func.py_func(x)
            assert np.allclose(r1, r2), (r1 - r2, r1.dtype, r2.dtype,
                                         func_name, x.dtype)

    if exceptions:
        raise Exception

if __name__ == '__main__':
    test_math_funcs()



# -*- coding: utf-8 -*-
"""
Miscellaneous (convenience) utilities.
"""
from __future__ import print_function, division, absolute_import

__all__ = ['ctypes', 'np', 'llvm', 'lc', 'MiniFunction']

try:
    import __builtin__ as builtins
except ImportError:
    import builtins

class UnavailableImport(object):
    def __init__(self, import_name):
        self.import_name = import_name

    def __getattr__(self, attr):
        __import__(self.import_name)

try:
    import ctypes
    have_ctypes = True
except ImportError:
    ctypes = UnavailableImport("ctypes")
    have_ctypes = False

try:
    import numpy as np
except ImportError:
    np = UnavailableImport("np")

try:
    import llvm.core
    from llvm import core as lc
except ImportError:
    llvm = UnavailableImport("llvm")
    lc = UnavailableImport("llvm.core")

from . import treepath

#
### Convenience utilities
#

def build_kernel_call(func_name, signature, miniargs, builder):
    """
    Call the kernel `lfunc` in a bunch of loops with scalar arguments.
    """
    # Build the kernel function signature
    funcname = builder.funcname(signature, func_name, is_external=False)

    # Generate 'lhs[i, j] = kernel(A[i, j], B[i, j])'
    lhs = miniargs[0].variable
    kernel_args = [arg.variable for arg in miniargs[1:]]
    funccall = builder.funccall(funcname, kernel_args, inline=True)
    assmt = builder.assign(lhs, funccall)
    if lhs.type.is_object:
        assmt = builder.stats(builder.decref(lhs), assmt)

    return assmt

def specialize(context, specializer_cls, ast, print_tree=False):
    "Specialize an AST with given specializer and compile"
    context = context or getcontext()
    specializers = [specializer_cls]
    result = next(iter(context.run(ast, specializers, print_tree=print_tree)))
    _, specialized_ast, _, code_result = result
    if not context.use_llvm:
        prototype, code_result = code_result
    return specialized_ast, code_result

class MiniFunction(object):
    """
    Convenience class to compile a function using LLVM and to invoke the
    function with ctypes given numpy arrays as input.
    """

    def __init__(self, context, specializer, variables, expr, name=None):
        self.b = context.astbuilder
        self.context = context
        self.specializer = specializer
        self.variables = variables
        self.minifunc = self.b.build_function(variables, expr, name)
        self.specialized_ast, (self.lfunc, self.ctypes_func) = specialize(
                                        context, specializer, self.minifunc)

    def get_ctypes_func_and_args(self, arrays):
        from .ctypes_conversion import get_data_pointer

        fist_array = arrays[0]
        shape = fist_array.shape
        for variable, array in zip(self.variables, arrays):
            for dim, extent in enumerate(array.shape):
                if extent != shape[dim] and extent != 1:
                    raise ValueError("Differing extents in dim %d (%s, %s)" %
                                     (dim, extent, shape[dim]))

        args = [fist_array.ctypes.shape]
        for variable, array in zip(self.variables, arrays):
            if variable.type.is_array:
                data_pointer = get_data_pointer(array, variable.type)
                args.append(data_pointer)
                if not self.specializer.is_contig_specializer:
                    args.append(array.ctypes.strides)
            else:
                raise NotImplementedError

        return args

    def __call__(self, *args, **kwargs):
        import numpy as np

        # print self.minifunc.ndim
        # self.minifunc.print_tree(self.context)
        # print self.context.debug_c(self.minifunc, self.specializer)

        out = kwargs.pop('out', None)
        assert not kwargs, kwargs

        if out is None:
            from . import minitypes
            dtype = minitypes.map_minitype_to_dtype(self.variables[0].type)
            broadcast = np.broadcast(*args)
            out = np.empty(broadcast.shape, dtype=dtype)

        arrays = [out]
        arrays.extend(args)
        assert len(arrays) == len(self.variables)

        args = self.get_ctypes_func_and_args(arrays)
        self.ctypes_func(*args)
        return out


def xpath(ast, expr):
    return treepath.find_all(ast, expr)


# Compatibility with Python 2.4
def any(it):
    for obj in it:
        if obj:
            return True
    return False

def all(it):
    for obj in it:
        if not obj:
            return False
    return True

class ComparableObjectMixin(object):
    "Make sure subclasses implement comparison and hashing methods"

    def __hash__(self):
        "Implement in subclasses"
        raise NotImplementedError

    def __eq__(self, other):
        "Implement in subclasses"
        return NotImplemented



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
# Use Python.asdl to test bad ast.

import unittest, ast, sys

from numba.asdl.tests import support

class TestBad(support.SchemaTestCase):
    def make_verification_callable(self, the_ast):
        def _verify():
            self.schema.verify(the_ast)
        return _verify

    def test_module_missing_body(self):
        the_ast = ast.Module()
        if sys.version_info < (2,7):
            self.assertRaises(support.schema.SchemaError,
                              self.make_verification_callable(the_ast))
        else:
            with self.capture_error() as cap:
                self.schema.verify(the_ast)
            self.assertEqual(str(cap.exception),
                             "At Module.body: Missing field")

    def test_missing_expr_context(self):
        a_name = ast.Name(id='x') # missing expr_context
        a_binop = ast.BinOp(left=a_name, right =a_name, op=ast.Add())
        the_ast = ast.Expr(value=a_binop)
        if sys.version_info < (2,7):
            self.assertRaises(support.schema.SchemaError,
                              self.make_verification_callable(the_ast))
        else:
            with self.capture_error() as cap:
                self.schema.verify(the_ast)
            self.assertEqual(str(cap.exception),
                             "At Name.ctx: Missing field")

    def test_wrong_arg(self):
        bad = ast.Raise() # doesn't matter what is inside
        args = ast.arguments(args=[bad], defaults=[])
        the_ast = ast.FunctionDef(name="haha",
                                  args=args,
                                  body=[],
                                  decorator_list=[])
        if sys.version_info < (2,7):
            self.assertRaises(support.schema.SchemaError,
                              self.make_verification_callable(the_ast))
        else:
            with self.capture_error() as cap:
                self.schema.verify(the_ast)
            self.assertIn(
                str(cap.exception),
                ["At arguments.args[0]: Cannot be a Raise",
                 "At arguments.args[0]: Expecting arg but got Raise"])

    def test_return_missing_lineno(self):
        the_ast = ast.Return(col_offset=0)
        if sys.version_info < (2,7):
            self.assertRaises(support.schema.SchemaError,
                              self.make_verification_callable(the_ast))
        else:
            with self.capture_error() as cap:
                self.schema.verify(the_ast)
            self.assertEqual(str(cap.exception),
                             "At Return.lineno: Missing field")


if __name__ == '__main__':
    unittest.main()




from numba import *



def simple_while(n):
    while n > 0:
        n -= 1
        a = 0
    return a

def simple_while_break(n):
    while n > 0:
        n -= 1
        break
    else:
        a = 1
    return a

def simple_while_pos(n):
    while n > 0:
        n -= 1
        a = 0
    else:
        a = 1
    return a

#def while_finally_continue(p, f):
#    while p():
#        try:
#            x = f()
#        finally:
#            print x
#            continue
#
#def while_finally_break(p, f):
#    while p():
#        try:
#            x = f()
#        finally:
#            print x
#            break
#
#def while_finally_outer(p, f):
#    x = 1
#    try:
#        while p():
#            print x
#            x = f()
#            if x > 0:
#                continue
#            if x < 0:
#                break
#    finally:
#        del x

def jitfunc(func):
    jit(int_(int_), warnstyle='simple')(func)

__doc__ = """
>>> jitfunc(simple_while)
Warning 9:11: local variable 'a' might be referenced before assignment
>>> jitfunc(simple_while_break)
Warning 17:11: local variable 'a' might be referenced before assignment
>>> jitfunc(simple_while_pos)
"""

if __name__ == "__main__":
#    jitfunc(simple_while_break)
    import numba
    numba.testing.testmod()



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
from numba.nodes import *

class TempNode(ExprNode): #, ast.Name):
    """
    Create a temporary to store values in. Does not perform reference counting.
    """

    temp_counter = 0

    def __init__(self, type, name=None, dst_variable=None):
        self.type = type
        self.name = name
        self.variable = Variable(type, name='___numba_%d' % self.temp_counter,
                                 is_local=True)
        TempNode.temp_counter += 1
        self.llvm_temp = None

        self.dst_variable = dst_variable
        self._tbaa_node = None

    def get_tbaa_node(self, tbaa):
        """
        TBAA metadata node unique to this temporary. This is valid
        since one cannot take the address of a temporary.
        """
        if self._tbaa_node is None:
            root = tbaa.get_metadata(char.pointer())
            self._tbaa_node = tbaa.make_unique_metadata(root)

        return self._tbaa_node

    def load(self, invariant=False):
        return TempLoadNode(temp=self, invariant=invariant)

    def store(self):
        return TempStoreNode(temp=self)

    def __repr__(self):
        if self.name:
            name = ", %s" % self.name
        else:
            name = ""
        return "temp(%s%s)" % (self.type, name)

class TempLoadNode(ExprNode):
    _fields = ['temp']

    def __init__(self, temp, invariant=False):
        self.temp = temp
        self.type = temp.type
        self.variable = Variable(self.type)
        self.invariant = invariant

    def __repr__(self):
        return "load(%s)" % self.temp

class TempStoreNode(TempLoadNode):
    _fields = ['temp']

    def __repr__(self):
        return "store(%s)" % self.temp



# Example from Travis Oliphant

import math
import numpy as np
from numba import jit, autojit
try:
    from numba.tests.support import ctypes_values as rng
except ImportError:
    rng = None

#@jit('double[:,:](int64, int64)')
@autojit
def gibbs(rk_seed, N, thin):
    rk_seed(0, rng.state_p)

    x = 0
    y = 0
    samples = np.empty((N,2))
    for i in range(N):
        for j in range(thin):
            #x = np.random.gamma(3,1.0/(y**2+4))
            x = rng.rk_gamma(rng.state_p, 3.0, 1.0/(y**2+4))
            #y = np.random.normal(1.0/(x+1), 1.0/math.sqrt(2+2*x))
            y = rng.rk_normal(rng.state_p, 1.0/(x+1), 1.0/math.sqrt(2+2*x))

        samples[i, 0] = x
        samples[i, 1] = y

    return samples

def test():
    if rng is not None:
        assert np.allclose(gibbs(rng.rk_seed, 10, 10),
                           gibbs.py_func(rng.rk_seed, 10, 10))

if __name__ == '__main__':
    test()



import numpy as np
import unittest

from numba import void, int32, uint32, jit, int64

@jit(void(uint32[:], uint32, uint32))
def prng(X, A, C):
    for i in range(X.shape[0]):
        for j in range(100):
            v = (A * X[i] + C)
            X[i] = v & 0xffffffff

@jit(uint32())
def unsigned_literal():
    return abs(0xFFFFFFFF)

@jit(int64())
def unsigned_literal_64():
    return 0x100000000

@jit(int64(int32))
def constant_int_add(a):
    return 0xffffffff + a

class Test(unittest.TestCase):
    def test_prng(self):
        N = 100
        A = 1664525
        C = 1013904223
        X0 = np.arange(N, dtype=np.uint32)
        X1 = X0.copy()
        prng.py_func(X0, A, C)
        prng(X1, A, C)
        self.assertTrue(np.all(X1 >= 0))
        self.assertTrue(np.all(X0 == X1))

    def test_unsigned_literal(self):
        got = unsigned_literal()
        expect = abs(0xFFFFFFFF)
        self.assertEqual(expect, got)

    def test_unsigned_literal_64(self):
        got = unsigned_literal_64()
        expect = 0x100000000
        self.assertEqual(expect, got)

    def test_constant_int_add(self):
        got = constant_int_add(1)
        expect = 0xffffffff + 1
        self.assertEqual(expect, got)

if __name__ == '__main__':
    unittest.main()



from numba.control_flow.tests.test_cfg_type_infer import *
from numba.testing.test_support import autojit_py3doc



@autojit_py3doc(warnstyle='simple')
def test_circular_error():
    """
    >>> test_circular_error()
    Traceback (most recent call last):
        ...
    NumbaError: Unable to infer type for assignment to ..., insert a cast or initialize the variable.
    """
    for i in range(10):
        if i > 5:
            var1 = var2
        else:
            var2 = var1

@autojit(warnstyle='simple')
def test_simple_circular():
    """
    >>> test_simple_circular()
    Warning 29:16: local variable 'y' might be referenced before assignment
    """
    x = 2.0
    for i in range(10):
        if i > 5:
            x = y
        else:
            y = x

@autojit(warnstyle='simple')
def test_simple_circular2():
    """
    >>> test_simple_circular2()
    Warning 44:16: local variable 'x' might be referenced before assignment
    """
    y = 2.0
    for i in range(10):
        if i > 5:
            x = y
        else:
            y = x

@autojit
def test_simple_circular3():
    """
    >>> test_simple_circular3()
    (Value(5), Value(5))
    >>> sig, syms = infer(test_simple_circular3.py_func,
    ...                   functype(None, []))
    >>> types(syms, 'x', 'y')
    (object_, object_)
    """
    x = values[5]
    y = 2.0
    for i in range(10):
        if i > 5:
            x = y
        else:
            y = x

    return x, y

@autojit
def test_simple_circular_promotion():
    """
    >>> test_simple_circular_promotion()
    ((3-3j), (1-3j))
    >>> sig, syms = infer(test_simple_circular_promotion.py_func,
    ...                   functype(None, []))
    >>> types(syms, 'x', 'y')
    (complex128, complex128)
    """
    x = 1
    y = 2
    for i in range(10):
        if i > 5:
            x = y + 2.0
        else:
            y = x - 3.0j

    return x, y

@autojit
def test_simple_circular_binop_promotion():
    """
    >>> test_simple_circular_binop_promotion()
    ((3-3j), (3+0j))
    >>> sig, syms = infer(test_simple_circular_binop_promotion.py_func,
    ...                   functype(None, []))
    >>> types(syms, 'x', 'y')
    (complex128, complex128)
    """
    x = 1
    y = 2
    for i in range(10):
        if i > 5:
            x = y - 3.0j
        else:
            y = x + 2.0 # In pure python, y would always be a float

    return x, y

#------------------------------------------------------------------------
# Test Unary/Binary Operations and Comparisons
#------------------------------------------------------------------------

@autojit_py3doc(warn=False)
def test_circular_binop():
    """
    >>> test_circular_binop()
    (1.0, 2.0, 1.0, -3L)
    >>> sig, syms = infer(test_circular_binop.py_func,
    ...                   functype(None, []), warn=False)
    >>> types(syms, 'x', 'y', 'z', 'a')
    (double, double, double, int)
    """
    x = 1
    y = 2
    for i in range(10):
        if i > 5:
            x = y - z
            z = 1.0
        else:
            z = int(x + y)
            y = x + z - y
            a = -z

    return x, y, z, a

@autojit(warn=False)
def test_circular_compare():
    """
    >>> test_circular_compare()
    (5.0, 1.0)
    >>> sig, syms = infer(test_circular_compare.py_func,
    ...                   functype(None, []), warn=False)
    >>> types(syms, 'x', 'y')
    (double, double)
    """
    x = 1
    for i in range(10):
        if i == 0:
            y = float(x)
        if x < 5:
            x += y

    return x, y

@autojit(warn=False)
def test_circular_compare2():
    """
    >>> test_circular_compare2()
    (2.0, 1.0)
    >>> sig, syms = infer(test_circular_compare.py_func,
    ...                   functype(None, []), warn=False)
    >>> types(syms, 'x', 'y')
    (double, double)
    """
    x = 1
    for i in range(10):
        if i == 0:
            y = float(x)
        if x < 5 and (x > 2 or i == 0):
            x += y

    return x, y

@autojit_py3doc(warn=False)
def test_circular_compare3():
    """
    >>> test_circular_compare3()
    1
    2
    3
    4
    (False, 10L)
    >>> sig, syms = infer(test_circular_compare3.py_func,
    ...                   functype(None, []), warn=False)
    >>> [str(x) for x in types(syms, 'cond', 'x')]
    ['bool', 'Py_ssize_t']
    """
    x = 1
    cond = True
    for i in range(10):
        if cond:
            x = i
        else:
            x = i + 1
        cond = x > 1 and x < 5
        if cond:
            x = cond or x < i
            cond = x
            x = i
            print(i)

    return cond, x

#------------------------------------------------------------------------
# Test Indexing
#------------------------------------------------------------------------

@autojit_py3doc(warn=False)
def test_delayed_array_indexing():
    """
    >>> test_delayed_array_indexing()
    (array([ 0.,  1.,  2.,  3.,  4.,  5.,  6.,  7.,  8.,  9.]), 1.0, 10L)
    >>> sig, syms = infer(test_delayed_array_indexing.py_func,
    ...                   functype(None, []), warn=False)
    >>> types(syms, 'array', 'var', 'x')
    (float64[:], float64, int)
    """
    array = np.ones(10, dtype=np.double)
    x = 0
    for i in range(11):
        var = array[x]
        array[x] = var * x
        x = int(i * 1.0)

    return array, var, x

@autojit(warn=False)
def test_delayed_array_slicing():
    """
    >>> array, row = test_delayed_array_slicing()
    >>> array2, row2 = test_delayed_array_slicing.py_func()
    >>> assert np.all(array == array2)
    >>> assert np.all(row == row2)

    >>> sig, syms = infer(test_delayed_array_slicing.py_func,
    ...                   functype(None, []), warn=False)
    >>> types(syms, 'array', 'row')
    (float64[:, :], float64[:])
    """
    array = np.ones((8, 10), dtype=np.double)
    for i in range(8):
        row = array[i, :]
        array[i, i] = row[i] * i
        array = array[:, :]

    return array, row


@autojit(warn=False)
def test_delayed_array_slicing2():
    """
    >>> array, row = test_delayed_array_slicing2()
    >>> array2, row2 = test_delayed_array_slicing2.py_func()
    >>> assert np.all(array == array2)
    >>> assert np.all(row == row2)

    >>> sig, syms = infer(test_delayed_array_slicing.py_func,
    ...                   functype(None, []), warn=False)
    >>> types(syms, 'array', 'row')
    (float64[:, :], float64[:])
    """
    for i in range(8):
        if i == 0:
            array = np.ones((8, 10), dtype=np.double)

        row = array[i, :]
        array[i, i] = row[i] * i
        array = array[:, :]

    return array, row

@autojit_py3doc(warn=False)
def test_delayed_string_indexing_simple():
    """
    >>> test_delayed_string_indexing_simple()
    ('eggs', 3L)
    >>> sig, syms = infer(test_delayed_string_indexing_simple.py_func,
    ...                   functype(None, []), warn=False)
    >>> types(syms, 's', 'x')
    (const char *, Py_ssize_t)
    """
    s = "spam ham eggs"
    for i in range(4):
        if i < 3:
            x = i + i

        s = s[x:]
        x = i

    return s[1:], x

@autojit_py3doc(warn=False)
def test_delayed_string_indexing():
    """
    >>> test_delayed_string_indexing()
    ('ham eggs', 3L)
    >>> sig, syms = infer(test_delayed_string_indexing.py_func,
    ...                   functype(None, []), warn=False)
    >>> types(syms, 's', 'x')
    (const char *, Py_ssize_t)
    """
    s = "spam ham eggs"
    for i in range(4):
        if i < 3:
            x = i
            tmp1 = s[x:]
            tmp2 = tmp1
            s = tmp2
        elif i < 5:
            s = tmp1[x:]
        else:
            s = "hello"

        x = i

    return s, x

@autojit_py3doc(warn=False)
def test_delayed_string_indexing2():
    """
    >>> test_delayed_string_indexing2()
    ('ham eggs', 3L)
    >>> sig, syms = infer(test_delayed_string_indexing2.py_func,
    ...                   functype(None, []), warn=False)
    >>> types(syms, 's', 'x')
    (const char *, Py_ssize_t)
    """
    for i in range(4):
        if i == 0:
            s = "spam ham eggs"

        if i < 3:
            x = i
            tmp1 = s[x:]
            tmp2 = tmp1
            s = tmp2
        elif i < 5:
            s = tmp1[x:]
        else:
            s = "hello"

        x = i

    return s, x

@autojit_py3doc(warn=False)
def test_string_indexing_error():
    """
    >>> test_string_indexing_error()
    Traceback (most recent call last):
        ...
    NumbaError: Cannot promote types (char, const char *) for variable s
    """
    for i in range(4):
        if i == 0:
            s = "spam ham eggs"

        if i < 3:
            s = s[i]
        elif i < 5:
            s = s[i:]

@autojit_py3doc(warn=False)
def test_string_indexing_error2():
    """
    >>> chr(test_string_indexing_error2())
    Traceback (most recent call last):
        ...
    NumbaError: Cannot promote types (char, const char *) for variable s
    """
    for i in range(4):
        if i == 0:
            s = "spam ham eggs"
        s = s[i]

    return s

@autojit(warn=False)
def test_string_indexing_valid():
    """
    >>> chr(test_string_indexing_valid())
    'm'
    """
    for i in range(4):
        s = "spam ham eggs"
        s = s[i]

    return s

#------------------------------------------------------------------------
# Test circular Calling of functions
#------------------------------------------------------------------------

@autojit
def simple_func(x):
    y = x * x + 4
    return y

@autojit_py3doc(warn=False)
def test_simple_call():
    """
    >>> test_simple_call()
    1091100052L
    >>> infer_simple(test_simple_call, 'x')
    (int,)
    """
    x = 0
    for i in range(10):
        x = simple_func(x)

    return x

@autojit
def func_with_promotion(x):
    y = x * x + 4.0
    return y

@autojit(warn=False)
def test_simple_call_promotion():
    """
    >>> test_simple_call_promotion()
    26640768404.0
    >>> infer_simple(test_simple_call_promotion, 'x')
    (double,)
    """
    x = 0
    for i in range(5):
        x = func_with_promotion(x)

    return x

#print test_simple_call_promotion.py_func()

@autojit
def func_with_promotion2(x):
    y = x * x + 4.0
    return np.sqrt(y) + 1j

@autojit(warn=False)
def test_simple_call_promotion2():
    """
    >>> result =test_simple_call_promotion2()
    >>> "%.4f" % round(result.real, 4)
    '3.9818'
    >>> round(result.imag, 4)
    3.9312
    >>> infer_simple(test_simple_call_promotion2, 'x')
    (complex128,)
    """
    x = 0
    for i in range(5):
        x = func_with_promotion2(x)

    return x

#print test_simple_call_promotion2.py_func()

#------------------------------------------------------------------------
# Test Utilities
#------------------------------------------------------------------------

def infer_simple(numba_func, *varnames):
    sig, syms = infer(numba_func.py_func, functype(None, []), warn=False)
    return types(syms, *varnames)

#from numba.minivect import minitypes
#sig = minitypes.FunctionType(None, [])
#func = jit(sig)(test_delayed_string_indexing_simple.py_func)
#print vars(func)
#test_delayed_string_indexing_simple()
#test_delayed_array_slicing2()
#test_simple_circular()
#test_circular_error()
testmod()



#! /usr/bin/env python
# ______________________________________________________________________

'''
Duplicated test for test_ast_getattr.py
'''

from numba.llvm_types import _plat_bits
from numba.decorators import jit

import numpy

import unittest

# ______________________________________________________________________

def get_ndarray_ndim(ndarr):
    return ndarr.ndim

def get_ndarray_shape(ndarr):
    return ndarr.shape

def get_ndarray_data(ndarr):
    return ndarr.data

def get_ndarray_2_shape_unpack_0(ndarr):
    dim0, _ = ndarr.shape
    return dim0

def get_ndarray_2_shape_unpack_1(ndarr):
    _, dim1 = ndarr.shape
    return dim1

# ______________________________________________________________________

class TestGetattr(unittest.TestCase):
    def test_getattr_ndim_1(self):
        test_data1 = numpy.array([1., 2., 3.])
        compiled_fn1 = jit('i(double[:])')(get_ndarray_ndim)
        self.assertEqual(compiled_fn1(test_data1), 1)

    def test_getattr_ndim_2(self):
        test_data2 = numpy.array([[1., 2., 3.], [4., 5., 6.]])
        compiled_fn2 = jit('i(double[:,:])')(get_ndarray_ndim)
        self.assertEqual(compiled_fn2(test_data2), 2)

    def test_getattr_shape_1(self):
        test_data = numpy.array([1., 2., 3.])
        compiled_fn = jit('i%d*(f8[:])' % (_plat_bits // 8))(get_ndarray_shape)
        result = compiled_fn(test_data)
        self.assertEqual(result[0], 3)

    def test_getattr_shape_2(self):
        test_data2 = numpy.array([[1., 2., 3.], [4., 5., 6.]])
        compiled_fn2 = jit('i%d*(f8[:])' % (_plat_bits // 8))(get_ndarray_shape)
        result = compiled_fn2(test_data2)
        self.assertEqual(result[0], 2)
        self.assertEqual(result[1], 3)

    def test_getattr_shape_2_unpack(self):
        compiler_fn = jit('i%d(d[:,:])' % (_plat_bits // 8))
        dim0_fn, dim1_fn = (compiler_fn(fn)
                            for fn in (get_ndarray_2_shape_unpack_0,
                                       get_ndarray_2_shape_unpack_1))
        test_data2 = numpy.array([[1., 2., 3.], [4., 5., 6.]])
        self.assertEqual(dim0_fn(test_data2), 2)
        self.assertEqual(dim1_fn(test_data2), 3)

    def test_getattr_data_1(self):
        test_data = numpy.array([1., 2., 3.])
        compiled_fn = jit('d*(d[:])')(get_ndarray_data)
        result = compiled_fn(test_data)
        self.assertEqual(result[0], 1.)
        self.assertEqual(result[1], 2.)
        self.assertEqual(result[2], 3.)

    def test_getattr_data_2(self):
        test_data = numpy.array([[1., 2., 3.], [4., 5., 6.]])
        compiled_fn = jit('d*(d[:,:])')(get_ndarray_data)
        result = compiled_fn(test_data)
        self.assertEqual(result[0], 1.)
        self.assertEqual(result[1], 2.)
        self.assertEqual(result[2], 3.)
        self.assertEqual(result[3], 4.)
        self.assertEqual(result[4], 5.)
        self.assertEqual(result[5], 6.)

# ______________________________________________________________________

if __name__ == "__main__":
    unittest.main()

# ______________________________________________________________________
# End of test_getattr.py



import numpy as np
from numba import int32, uint32, float32, float64
from numba.vectorize import Vectorize
from timeit import default_timer as time
import unittest

def vector_add(a, b):
    return a + b

class TestBasicVectorize(unittest.TestCase):
    def setUp(self):
        bv = Vectorize(vector_add, backend='ast')
        bv.add(restype=int32,   argtypes=[int32,    int32])
        bv.add(restype=uint32,  argtypes=[uint32,   uint32])
        bv.add(restype=float32, argtypes=[float32,  float32])
        bv.add(restype=float64, argtypes=[float64, 	float64])
        self.basic_ufunc = bv.build_ufunc()

    def _test(self, ty):
        print("Test %s" % ty)
        data = np.linspace(0., 10000., 100000).astype(ty)

        ts = time()
        result = self.basic_ufunc(data, data)
        tnumba = time() - ts

        ts = time()
        gold = np.add(data, data)
        tnumpy = time() - ts

        print("Numpy time: %fs" % tnumpy)
        print("Numba time: %fs" % tnumba)

        if tnumba < tnumpy:
            print("Numba is FASTER by %fx" % (tnumpy/tnumba))
        else:
            print("Numba is SLOWER by %fx" % (tnumba/tnumpy))


        for expect, got in zip(gold, result):
            assert expect == got

    def test_double(self):
        self._test(np.double)

    def test_float(self):
        self._test(np.float)

    def test_int32(self):
        self._test(np.int32)

    def test_uint32(self):
        self._test(np.uint32)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import llvm.core

from numba import *
from numba.external import external
from numba.external.utilities import utilities
from numba.exttypes.virtual import PyCustomSlots_Table

class UtilityFunction(external.ExternalFunction):
    """
    A utility function written in a native language.

    funcaddr: the integer address of the C utility function
    See ExternalFunction for keyword arguments!
    """

    def __init__(self, funcaddr, return_type, arg_types, **kwargs):
        super(UtilityFunction, self).__init__(return_type, arg_types, **kwargs)
        self.funcaddr = funcaddr

    def declare_lfunc(self, context, llvm_module):
        lsig = self.signature.pointer().to_llvm(context)
        inttype = Py_uintptr_t.to_llvm(context)
        intval = llvm.core.Constant.int(inttype, self.funcaddr)
        lfunc = intval.inttoptr(lsig)
        return lfunc

    @classmethod
    def load(cls, func_name, signature, **kwds):
        """
        Load a utility function by name from the
        numba.external.utilities.utilities module.
        """
        # Get the integer address of C utility function
        func_addr = getattr(utilities, func_name)
        return cls(func_addr, signature.return_type, signature.args,
                   func_name=func_name, **kwds)


load = UtilityFunction.load
load2 = lambda name, sig: load(name, sig, check_pyerr_occurred=True)

object_to_numeric = {
    char       : load2("__Numba_PyInt_AsSignedChar", char(object_)),
    uchar      : load2("__Numba_PyInt_AsUnsignedChar", uchar(object_)),
    short      : load2("__Numba_PyInt_AsSignedShort", short(object_)),
    ushort     : load2("__Numba_PyInt_AsUnsignedShort", ushort(object_)),
    int_       : load2("__Numba_PyInt_AsSignedInt", int_(object_)),
    uint       : load2("__Numba_PyInt_AsUnsignedInt", uint(object_)),
    long_      : load2("__Numba_PyInt_AsSignedLong", long_(object_)),
    ulong      : load2("__Numba_PyInt_AsUnsignedLong", ulong(object_)),
    longlong   : load2("__Numba_PyInt_AsSignedLongLong", longlong(object_)),
    ulonglong  : load2("__Numba_PyInt_AsUnsignedLongLong", ulonglong(object_)),
}

void_p = void.pointer()
void_pp = void_p.pointer()

utility_funcs = list(object_to_numeric.itervalues()) + [
    UtilityFunction.load(
        "lookup_method", void_p(void_pp, uint64, char.pointer())),
]

def default_utility_library(context):
    """
    Create a library of utility functions.
    """
    extlib = external.ExternalLibrary(context)

    for utility_func in utility_funcs:
        extlib.add(utility_func)

    return extlib



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import numpy as np

from numba import decorators
from numba.codegen.llvmcontext import LLVMContextManager
from numba.minivect import minitypes
from . import _internal

try:
    ptr_t = long
except NameError:
    ptr_t = int
    assert False, "Have not check this yet" # Py3.0?

_llvm_ty_str_to_numpy = {
            'i8'     : np.int8,
            'i16'    : np.int16,
            'i32'    : np.int32,
            'i64'    : np.int64,
            'float'  : np.float32,
            'double' : np.float64,
        }

def _llvm_ty_to_numpy(ty):
    return _llvm_ty_str_to_numpy[str(ty)]

def _llvm_ty_to_dtype(ty):
    return np.dtype(_llvm_ty_to_numpy(ty)) #.num

_numbatypes_str_to_numpy = {
            'int8'     : np.int8,
            'int16'    : np.int16,
            'int32'    : np.int32,
            'int64'    : np.int64,
            'uint8'    : np.uint8,
            'uint16'   : np.uint16,
            'uint32'   : np.uint32,
            'uint64'   : np.uint64,
#            'f'        : np.float32,
#            'd'        : np.float64,
            'float'    : np.float32,
            'double'   : np.float64,
        }

def _numbatypes_to_numpy(ty):
    # ret = _numbatypes_str_to_numpy[str(ty)]
    return ty.get_dtype()

class CommonVectorizeFromFunc(object):
    def build(self, lfunc, dtypes):
        raise NotImplementedError

    def get_dtype_nums(self, tyslist):
        return [[dtype.num for dtype in dtypes] for dtypes in tyslist]

    def __call__(self, lfunclist, tyslist, engine,
                 dispatcher=None,
                 **kws):
        '''create ufunc from a llvm.core.Function

        lfunclist : a single or iterable of llvm.core.Function instance
        engine : a llvm.ee.ExecutionEngine instance

        return a function object which can be called from python.
        '''
        try:
            iter(lfunclist)
        except TypeError:
            lfunclist = [lfunclist]

        ptrlist = self._prepare_pointers(lfunclist, tyslist, engine, **kws)

        fntype = lfunclist[0].type.pointee
        inct = len(fntype.args)
        outct = 1

        datlist = [None] * len(lfunclist)

        # Becareful that fromfunc does not provide full error checking yet.
        # If typenum is out-of-bound, we have nasty memory corruptions.
        # For instance, -1 for typenum will cause segfault.
        # If elements of type-list (2nd arg) is tuple instead,
        # there will also memory corruption. (Seems like code rewrite.)
        tyslist = self.get_dtype_nums(tyslist)
        ufunc = _internal.fromfunc(ptrlist, tyslist, inct, outct,
                                   datlist, dispatcher)
        return ufunc

    def _prepare_ufunc_core(self, lfunclist, tyslist, **kws):
        spuflist = []
        for i, (lfunc, dtypes) in enumerate(zip(lfunclist, tyslist)):
            spuflist.append(self.build(lfunc, dtypes, **kws))
        return spuflist

    def _get_pointer_from_ufunc_core(self, spuf, engine):
        fptr = engine.get_pointer_to_function(spuf)
        return ptr_t(fptr)

    def _prepare_pointers(self, lfunclist, tyslist, engine, **kws):
        spuflist = self._prepare_ufunc_core(lfunclist, tyslist, **kws)
        ptrlist = [self._get_pointer_from_ufunc_core(spuf, engine)
                   for spuf in spuflist]
        return ptrlist

    def _prepare_prototypes_and_pointers(self, lfunclist, tyslist, engine, **kws):
        spuflist = self._prepare_ufunc_core(lfunclist, tyslist, **kws)
        ptrlist = [self._get_pointer_from_ufunc_core(spuf, engine)
                   for spuf in spuflist]
        return zip(spuflist, ptrlist)

class GenericASTVectorize(object):

    def __init__(self, func):
        self.pyfunc = func
        self.translates = []
        self.args_restypes = getattr(self, 'args_restypes', [])
        self.signatures = []
        self.llvm_context = LLVMContextManager()

    def _get_tys_list(self):
        types_lists = []
        for numba_func in self.translates:
            dtype_nums = []
            types_lists.append(dtype_nums)
            for arg_type in self.get_argtypes(numba_func):
                dtype = minitypes.map_minitype_to_dtype(arg_type)
                dtype_nums.append(dtype)

        return types_lists

    def _get_lfunc_list(self):
        return [t.lfunc for t in self.translates]

    def _get_ee(self):
        return self.llvm_context.execution_engine

    def build_ufunc(self):
        raise NotImplementedError

    def register_ufunc(self, ufunc):
        from numba.type_inference.modules import numpyufuncs
        numpyufuncs.register_arbitrary_ufunc(ufunc)

    def _from_func_factory(self, lfunclist, tyslist, **kws):
        """
        Set this attribute to some subclass of CommonVectorizeFromFunc
        """
        raise NotImplementedError

    def _from_func(self, **kws):
        assert self.translates, "No translation"
        lfunclist = self._get_lfunc_list()
        tyslist = self._get_tys_list()
        engine = self._get_ee()
        ufunc = self._from_func_factory(lfunclist, tyslist, engine=engine, **kws)
        self.register_ufunc(ufunc)
        return ufunc

    def build_ufunc_core(self, **kws):
        '''Build the ufunc core functions and returns the prototype and pointer.
        The return value is a list of tuples (prototype, pointer).
        '''
        assert self.translates, "No translation"
        lfunclist = self._get_lfunc_list()
        tyslist = self._get_tys_list()
        engine = self._get_ee()
        get_proto_ptr = self._from_func_factory._prepare_prototypes_and_pointers
        return get_proto_ptr(lfunclist, tyslist, engine, **kws)

    def add(self, restype=None, argtypes=None, **kwds):
        """
        Add a specialization to the vectorizer. Pass any keyword arguments
        to numba.jit().
        """
        dec = decorators.jit(restype, argtypes, **kwds)
        numba_func = dec(self.pyfunc)
        self.args_restypes.append(list(numba_func.signature.args) +
                                  [numba_func.signature.return_type])
        self.signatures.append((restype, argtypes, {}))
        self.translates.append(numba_func)

    def get_argtypes(self, numba_func):
        return list(numba_func.signature.args) + [numba_func.signature.return_type]


def post_vectorize_optimize(func):
    '''Perform aggressive optimization after each vectorizer.
        
    TODO: review if this is necessary.
    '''
    cm = LLVMContextManager()
    return cm.pass_manager


